# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020-2021 grommunio GmbH

from . import DB, OptionalC, OptionalNC, NotifyTable, logger
from services import Service
from tools import formats
from tools.constants import PropTags, PropTypes
from tools.DataModel import DataModel, Id, Text, Int, BoolP, RefProp, Bool, Date
from tools.DataModel import InvalidAttributeError, MismatchROError, MissingRequiredAttributeError
from tools.rop import nxTime

from sqlalchemy import Column, ForeignKey, event, func, inspect, select
from sqlalchemy.dialects.mysql import ENUM, INTEGER, TEXT, TIMESTAMP, TINYINT, VARBINARY, VARCHAR
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import column_property, relationship, selectinload, validates

import crypt
import json
import sys

from datetime import datetime

# Was moved in SQLA 1.4
import sqlalchemy
if sqlalchemy.__version__.split(".") >= ["1", "4"]:
    from sqlalchemy.orm import DeclarativeMeta
else:
    from sqlalchemy.ext.declarative.api import DeclarativeMeta


class Users(DataModel, DB.Base, NotifyTable):
    class PropMap():
        def __init__(self, user):
            self.__user = user
            self.__struct = {}
            self.__dict = {}
            for prop in user._properties:
                if PropTypes.ismv(prop.tag):
                    if prop.tag in self.__struct:
                        self.__dict[self._name(prop.tag)].append(prop.val)
                        self.__struct[prop.tag].append(prop)
                    else:
                        self.__dict[self._name(prop.tag)] = [prop.val]
                        self.__struct[prop.tag] = [prop]
                else:
                    self.__dict[self._name(prop.tag)] = prop.val
                    self.__struct[prop.tag] = prop

        @staticmethod
        def _name(key):
            return key.lower() if isinstance(key, str) else PropTags.lookup(key, hex(key)).lower()

        def __contains__(self, o):
            return self._name(o) in self.__dict

        def __getitem__(self, k):
            return self.__dict[self._name(k)]

        def __len__(self):
            return len(self.__dict)

        def __repr__(self):
            return repr(self.__dict)

        def __setitem__(self, k, v):
            tag = PropTags.deriveTag(k)
            name = self._name(k)
            if not PropTypes.ismv(tag):
                if tag in self.__struct:
                    if v is None:
                        DB.session.delete(self.__struct[tag])
                    else:
                        self.__struct[tag].val = v
                elif v is None:
                    return
                else:
                    self.__struct[tag] = UserProperties(tag, v, self.__user)
                    DB.session.add(self.__struct[tag])
                self.__dict[name] = v
                return
            if v is None:
                v = []
            elif not isinstance(v, (list, tuple, set)):
                v = [v]
            values = self.__dict.get(name, ())
            current = self.__struct.get(tag, ())
            next = []
            for value in v:
                if v in current:
                    i = values.index(v)
                    values.pop(i)
                    next.append(current.pop(i))
                else:
                    next.append(UserProperties(tag, value, self.__user))
                    DB.session.add(next[-1])
            for rm in current:
                DB.session.delete(rm)
            order = 1
            for up in next:
                up.orderID = order
                order += 1
            self.__dict[name] = v
            self.__struct[tag] = next

        def get(self, k, d=None):
            return self.__dict.get(self._name(k), d)

        def idmap(self):
            return {tag: [v.val for v in value] if PropTypes.ismv(tag) else value.val for tag, value in self.__struct.items()}

        def namemap(self):
            return self.__dict

        def items(self):
            return self.__dict.items()

        def update(self, data):
            for k, v in data.items():
                self[k] = v

        def rawmap(self):
            def getv(prop):
                return PropTypes.pyType(prop.baseType)(prop.content)
            return {tag: [getv(p) for p in prop] if PropTypes.ismv(tag) else getv(prop)
                    for tag, prop in self.__struct.items()}

    __tablename__ = "users"

    ID = Column("id", INTEGER(10, unsigned=True), nullable=False, primary_key=True, unique=True)
    username = Column("username", VARCHAR(320, charset="ascii"), nullable=False, unique=True)
    altname = OptionalC(111, "NULL", Column("altname", VARCHAR(64), unique=True))
    _password = Column("password", VARCHAR(40), nullable=False, server_default="")
    domainID = Column("domain_id", INTEGER(10, unsigned=True), nullable=False, index=True)
    maildir = Column("maildir", VARCHAR(128), nullable=False, server_default="")
    addressStatus = Column("address_status", TINYINT, nullable=False, server_default="0")
    privilegeBits = Column("privilege_bits", INTEGER(10, unsigned=True), nullable=False, default=0)
    externID = Column("externid", VARBINARY(64))
    chatID = OptionalC(78, "NULL", Column("chat_id", VARCHAR(26)))
    lang = Column("lang", VARCHAR(32), nullable=False, server_default="")
    homeserverID = OptionalC(104, "0", Column("homeserver", TINYINT(unsigned=True), nullable=False, server_default="0"))
    _syncPolicy = OptionalC(76, "NULL", Column("sync_policy", TEXT))
    _deprecated_maxSize = OptionalC(-111, "0", Column("max_size", INTEGER(10), nullable=False, default=0))
    _deprecated_addressType = OptionalC(-86, "NULL", Column("address_type", TINYINT, nullable=False, server_default="0"))
    _deprecated_subType = OptionalC(-85, "NULL", Column("sub_type", TINYINT, nullable=False, server_default="0"))
    _deprecated_groupID = Column("group_id", INTEGER(10, unsigned=True), nullable=False, index=True, default=0)

    domain = relationship("Domains", foreign_keys=domainID, primaryjoin="Users.domainID == Domains.ID")
    roles = relationship("AdminRoles", secondary="admin_user_role_relation", viewonly=True)
    _properties = relationship("UserProperties", cascade="all, delete-orphan", single_parent=True, passive_deletes=True,
                               order_by="UserProperties.orderID", back_populates="user")
    aliases = relationship("Aliases", cascade="all, delete-orphan", single_parent=True, passive_deletes=True,
                           back_populates="main")
    fetchmail = OptionalNC(75, [],
                           relationship("Fetchmail", cascade="all, delete-orphan", single_parent=True,
                                        order_by="Fetchmail.active.desc()", back_populates="user"))
    forward = relationship("Forwards", uselist=False, cascade="all, delete-orphan", back_populates="user")
    homeserver = OptionalNC(104, None,
                            relationship("Servers", foreign_keys=homeserverID, primaryjoin="Users.homeserverID == Servers.ID"))
    mlist = relationship("MLists", back_populates="user", uselist=False)

    _dictmapping_ = ((Id(), Text("username", flags="patch")),
                     (Id("domainID", flags="init"),
                      {"attr": "ldapID", "flags": "patch"},
                      Int("status", flags="patch"),
                      Text("altname", flags="patch")),
                     (Text("lang", match=False, flags="patch"),
                      BoolP("pop3_imap", flags="patch"),
                      BoolP("smtp", flags="patch"),
                      BoolP("changePassword", flags="patch"),
                      BoolP("publicAddress", flags="patch"),
                      BoolP("privChat", flags="patch"),
                      BoolP("privVideo", flags="patch"),
                      BoolP("privFiles", flags="patch"),
                      BoolP("privArchive", flags="patch"),
                      RefProp("aliases", flags="patch, managed", link="aliasname", flat="aliasname", qopt=selectinload),
                      RefProp("fetchmail", flags="managed, patch", link="ID", qopt=selectinload),
                      {"attr": "properties", "flags": "patch", "func": lambda p: p.namemap()},
                      RefProp("roles", qopt=selectinload),
                      RefProp("forward", flags="managed, patch"),
                      RefProp("mlist", filter="set", flat="ID"),
                      {"attr": "syncPolicy", "flags": "patch"},
                      {"attr": "chat", "flags": "patch"},
                      {"attr": "chatAdmin", "flags": "patch"},
                      Int("orgID"),
                      RefProp("homeserver", "homeserverID", flags="patch", filter="set", qopt=selectinload)),
                     ({"attr": "password", "flags": "init, hidden"},
                      Text("maildir", match=False, flags="hidden"),))

    USER_PRIVILEGE_POP3_IMAP = 1 << 0
    USER_PRIVILEGE_SMTP = 1 << 1
    USER_PRIVILEGE_CHGPASSWD = 1 << 2
    USER_PRIVILEGE_PUBADDR = 1 << 3
    USER_PRIVILEGE_CHAT = 1 << 4
    USER_PRIVILEGE_VIDEO = 1 << 5
    USER_PRIVILEGE_FILES = 1 << 6
    USER_PRIVILEGE_ARCHIVE = 1 << 7

    MAILUSER = 0x0
    DISTLIST = 0x1
    FORUM = 0x2
    AGENT = 0x3
    ORGANIZATION = 0x4
    PRIVATE_DISTLIST = 0x5
    REMOTE_MAILUSER = 0x6
    ROOM = 0x7
    EQUIPMENT = 0x8
    SEC_DISTLIST = 0x9
    CONTAINER = 0x100
    TEMPLATE = 0x101
    ADDRESS_TEMPLATE = 0x102
    SEARCH = 0x200

    NORMAL = 0
    SUSPENDED = 1
    OUTOFDATE = 2
    DELETED = 3
    SHARED = 4
    CONTACT = 5
    USER_MASK = 0x07
    DOMAIN_MASK = 0x30

    _chatUser = None
    _propcache = None

    @staticmethod
    def checkCreateParams(data):
        from orm.domains import Domains
        from tools.license import getLicense
        if "username" not in data:
            return "Missing username"
        if data.get("status", 0) not in (Users.SHARED, Users.CONTACT) and Users.count() >= getLicense().users:
            return "License user limit exceeded"
        if "domainID" in data:
            domain = Domains.query.filter(Domains.ID == data.get("domainID")).first()
        elif "@" in data["username"]:
            domain = Domains.query.filter(Domains.domainname == data["username"].split("@")[1]).first()
        else:
            domain = None
        if domain is None:
            return "Invalid domain"
        if data.get("status") != Users.CONTACT:
            if "@" in data["username"]:
                dname = data["username"].split("@", 1)[1]
                if domain.domainname != dname and domain.displayname != dname:
                    return "Domain specifications do not match"
            else:
                data["username"] += "@"+domain.domainname
        data["domain"] = domain
        data["domainID"] = domain.ID
        domainUsers = Users.count(Users.domainID == domain.ID)
        if domain.maxUser <= domainUsers:
            return "Maximum number of domain users reached"
        data["domainStatus"] = domain.domainStatus
        if "properties" not in data:
            data["properties"] = {}
        properties = data["properties"]
        properties["creationtime"] = datetime.now()
        if "displaytypeex" not in properties:
            properties["displaytypeex"] = 0

    def __init__(self, props, *args, **kwargs):
        self._permissions = None
        if props is None:
            return
        status = props.pop("domainStatus") << 4
        self.fromdict(props, *args, **kwargs)
        self.addressStatus = (self.addressStatus or 0) | status

    def fromdict(self, patches, syncStore=False, *args, **kwargs):
        isContact = patches.get("status", self.status) != Users.CONTACT
        if "username" in patches and patches["username"] != self.username:
            from orm.domains import Domains
            username = patches.pop("username")
            domain = patches.pop("domain", None) or Domains.query.filter(Domains.ID == self.domainID).first()
            if "@" in username:
                uname, dname = username.split("@", 1)
                if dname != domain.domainname and dname != domain.displayname:
                    raise ValueError("Domain specifications mismatch.")
                self.username = uname+"@"+domain.domainname
            elif not isContact:
                self.username = username+"@"+domain.domainname
            else:
                self.username = username
            if not isContact and not formats.email.match(self.username):
                raise ValueError("'{}' is not a valid e-mail address".format(self.username))
        DataModel.fromdict(self, patches, args, kwargs)
        displaytype = self.properties.get("displaytypeex", 0)
        if displaytype in (0, 1, 7, 8):
            self._deprecated_addressType, self._deprecated_subType = self._decodeDisplayType(displaytype)
        if self.chatID:
            with Service("chat", errors=Service.SUPPRESS_INOP) as chat:
                self._chatUser = chat.updateUser(self, False)
        if syncStore == "always" or (syncStore and "properties" in patches):
            self.syncStore()

    @staticmethod
    def _decodeDisplayType(displaytype):
        if displaytype == 0:
            return 0, 0
        elif displaytype == 1:
            return 2, 0
        elif displaytype == 7:
            return 0, 1
        elif displaytype == 8:
            return 0, 2
        raise ValueError("Unknown display type "+str(displaytype))

    def baseName(self):
        return self.username.rsplit("@", 1)[0]

    def domainName(self):
        return self.username.rsplit("@", 1)[1] if "@" in self.username else None

    def permissions(self):
        if self.ID == 0:
            from tools.permissions import Permissions
            return Permissions.sysadmin()
        if not hasattr(self, "_permissions") or self._permissions is None:
            from .roles import AdminUserRoleRelation as AURR, AdminRolePermissionRelation as ARPR, AdminRoles as AR
            from tools.permissions import Permissions
            perms = ARPR.query.filter(AURR.userID == self.ID).join(AR).join(AURR).all()
            self._permissions = Permissions.fromDB(perms)
        return self._permissions

    def getProp(self, name):
        return self.properties[name].val if name in self.properties else None

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, pw):
        # On OpenBSD only blowfish is supported
        if sys.platform.startswith("openbsd"):
            _method = crypt.METHOD_BLOWFISH
        else:
            _method = crypt.METHOD_SHA512
        self._password = crypt.crypt(pw, crypt.mksalt(_method))

    def chkPw(self, pw):
        return crypt.crypt(pw, self.password) == self.password

    @property
    def propmap_id(self):
        if self._propcache is None:
            self._propcache = self.PropMap(self)
        return self._propcache.idmap()

    @property
    def properties(self):
        if self._propcache is None:
            self._propcache = self.PropMap(self)
        return self._propcache

    @properties.setter
    def properties(self, values):
        if self._propcache is None:
            self._propcache = self.PropMap(self)
        self._propcache.update(values)

    @property
    def syncPolicy(self):
        try:
            return json.loads(self._syncPolicy)
        except Exception:
            return None

    @syncPolicy.setter
    def syncPolicy(self, value):
        self._syncPolicy = json.dumps(value, separators=(",", ":")) if value is not None else None

    def _setPB(self, bit, val):
        self.privilegeBits = (self.privilegeBits or 0) | bit if val else (self.privilegeBits or 0) & ~bit

    def _getPB(self, bit):
        return self.privilegeBits if isinstance(self, DeclarativeMeta) else bool((self.privilegeBits or 0) & bit)

    @hybrid_property
    def pop3_imap(self):
        return self._getPB(self.USER_PRIVILEGE_POP3_IMAP)

    @pop3_imap.setter
    def pop3_imap(self, val):
        self._setPB(self.USER_PRIVILEGE_POP3_IMAP, val)

    @pop3_imap.expression
    def pop3_imap(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_POP3_IMAP) != 0

    @hybrid_property
    def smtp(self):
        return self._getPB(self.USER_PRIVILEGE_SMTP)

    @smtp.setter
    def smtp(self, val):
        self._setPB(self.USER_PRIVILEGE_SMTP, val)

    @smtp.expression
    def smtp(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_SMTP) != 0

    @hybrid_property
    def changePassword(self):
        return self._getPB(self.USER_PRIVILEGE_CHGPASSWD)

    @changePassword.setter
    def changePassword(self, val):
        self._setPB(self.USER_PRIVILEGE_CHGPASSWD, val)

    @changePassword.expression
    def changePassword(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_CHGPASSWD) != 0

    @hybrid_property
    def publicAddress(self):
        return self._getPB(self.USER_PRIVILEGE_PUBADDR)

    @publicAddress.setter
    def publicAddress(self, val):
        self._setPB(self.USER_PRIVILEGE_PUBADDR, val)

    @publicAddress.expression
    def publicAddress(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_PUBADDR) != 0

    @hybrid_property
    def privChat(self):
        return self._getPB(self.USER_PRIVILEGE_CHAT)

    @privChat.setter
    def privChat(self, val):
        self._setPB(self.USER_PRIVILEGE_CHAT, val)

    @privChat.expression
    def privChat(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_CHAT) != 0

    @hybrid_property
    def privVideo(self):
        return self._getPB(self.USER_PRIVILEGE_VIDEO)

    @privVideo.setter
    def privVideo(self, val):
        self._setPB(self.USER_PRIVILEGE_VIDEO, val)

    @privVideo.expression
    def privVideo(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_VIDEO) != 0

    @hybrid_property
    def privFiles(self):
        return self._getPB(self.USER_PRIVILEGE_FILES)

    @privFiles.setter
    def privFiles(self, val):
        self._setPB(self.USER_PRIVILEGE_FILES, val)

    @privFiles.expression
    def privFiles(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_FILES) != 0

    @hybrid_property
    def privArchive(self):
        return self._getPB(self.USER_PRIVILEGE_ARCHIVE)

    @privArchive.setter
    def privArchive(self, val):
        self._setPB(self.USER_PRIVILEGE_ARCHIVE, val)

    @privArchive.expression
    def privArchive(cls):
        return cls.privilegeBits.op("&")(cls.USER_PRIVILEGE_ARCHIVE) != 0

    @property
    def ldapID(self):
        from services.ldap import LdapService
        return None if self.externID is None else LdapService.escape_filter_chars(self.externID)

    @ldapID.setter
    def ldapID(self, value):
        from services.ldap import LdapService
        self.externID = None if value is None else LdapService.unescapeFilterChars(value)

    @hybrid_property
    def status(self):
        return (self.addressStatus or 0) & self.USER_MASK

    @status.setter
    def status(self, val):
        self.addressStatus = ((self.addressStatus or 0) & ~self.USER_MASK) | (val & self.USER_MASK)

    @status.expression
    def status(cls):
        return cls.addressStatus.op("&")(cls.USER_MASK)

    @hybrid_property
    def domainStatus(self):
        return (self.addressStatus & self.DOMAIN_MASK) >> 4

    @domainStatus.setter
    def domainStatus(self, val):
        self.addressStatus = (self.addressStatus & ~self.DOMAIN_MASK) | (val << 4 & self.DOMAIN_MASK)

    @domainStatus.expression
    def domainStatus(cls):
        return cls.addressStatus.op("&")(cls.DOMAIN_MASK).op(">>")(4)

    @property
    def chat(self):
        if not self.chatID:
            return False
        if self._chatUser is None:
            with Service("chat", errors=Service.SUPPRESS_INOP) as chat:
                self._chatUser = chat.getUser(self.chatID)
        return self.domain.chat and self._chatUser["delete_at"] == 0 if self._chatUser else False

    @chat.setter
    def chat(self, value):
        if value == self.chat:
            return
        if not DB.minVersion(78):
            raise ValueError("Cannot activate chat - please upgrade database schema to at least 78")
        if value and self.addressStatus:
            raise ValueError("Cannot activate chat for locked user")
        err_prefix = "Could not enable chat for user '{}': ".format(self.username)
        if not self.domain.chat:
            logger.warning(err_prefix+"chat is not enabled for domain")
            return
        with Service("chat") as chat:
            if isinstance(value, str):
                tmp = chat.getUser(value)
                if tmp is None:
                    logger.warning(err_prefix+"chat user not found")
                    return
                self.chatID = value
                self._chatUser = tmp
                return
            if self._chatUser:
                tmp = chat.activateUser(self, value)
                if tmp:
                    self._chatUser["delete_at"] = not value
                err = "Failed to "+("" if value else "de")+"activate chat user"
            else:
                tmp = chat.createUser(self)
                if tmp:
                    self._chatUser = tmp
                err = err_prefix+"Failed to create user"
            if tmp is None:
                logger.warning(err)

    @property
    def chatAdmin(self):
        return self.chat and "system_admin" in self._chatUser["roles"].split(" ")

    @chatAdmin.setter
    def chatAdmin(self, value):
        if not self.chat or self.chatAdmin == bool(value):
            return
        if value:
            tmpRoles = " ".join(self._chatUser["roles"].split(" ")+["system_admin"])
        else:
            tmpRoles = " ".join(role for role in self._chatUser["roles"].split(" ") if role != "system_admin")
        with Service("chat") as chat:
            tmp = chat.setUserRoles(self.chatID, tmpRoles)
        if tmp is None:
            logger.warning("Failed to update chat user")
        self._chatUser["roles"] = tmpRoles

    @validates("_syncPolicy")
    def triggerSyncPolicyUpdate(self, key, value, *args):
        if value != self._syncPolicy:
            with Service("redis", errors=Service.SUPPRESS_INOP) as r:
                r.delete("grommunio-sync:policycache-"+self.username)
        return value

    @validates("homeserverID")
    def checkHomeserver(self, key, value, *args):
        from tools.config import Config
        if self.homeserverID and value != self.homeserverID and Config["options"].get("serverExplicitMount"):
            raise ValueError("Cannot change homeserver with explicitly mounted home-directories")
        from .misc import Servers
        if value and Servers.query.filter(Servers.ID == value).count() == 0:
            raise ValueError("Invalid homeserver")
        return value or 0

    @staticmethod
    def count(*filters):
        """Count users.

        Applies filters to only count real users (DISPLAYTYPEEX == 0) and ignore the admin user (ID == 0).

        Parameters
        ----------
        filters : iterable, optional
            Additional filter expressions to use. The default is ().

        Returns
        -------
        int
            Number of users
        """
        return Users.query.with_entities(Users.ID)\
                          .filter(Users.ID != 0, Users.maildir != "", Users.status != Users.SHARED, *filters)\
                          .count()

    def delete(self, deleteChatUser=True):
        """Delete user from database.

        Also cleans up entries in forwards and associations tables.

        Parameters
        ----------
        filters : iterable, optional
            Additional filter expressions to use. The default is ().

        Returns
        -------
        str
            Error message or None if successful.
        """
        from .mlists import Associations, MLists
        if self.ID == 0:
            raise ValueError("Cannot delete superuser")
        Forwards.query.filter(Forwards.username == self.username).delete(synchronize_session=False)
        Associations.query.filter(Associations.username == self.username).delete(synchronize_session=False)
        MLists.query.filter(MLists.listname == self.username).delete(synchronize_session=False)
        if self.chatID is not None:
            from services import Service
            with Service("chat", errors=Service.SUPPRESS_ALL) as chat:
                chat.deleteUser(self) if deleteChatUser else chat.activateUser(self, False)
        DB.session.delete(self)

    @staticmethod
    def create(props, externID=None, sync=True, *args, **kwargs):
        from .misc import Servers
        from tools.misc import AutoClean
        from tools.storage import UserSetup
        error = Users.checkCreateParams(props)
        chat = props.pop("chat", None)
        if error is not None:
            return error, 400
        try:
            user = Users(props)
            user.externID = externID
        except (InvalidAttributeError, MismatchROError, MissingRequiredAttributeError, ValueError) as err:
            return err.args[0], 400
        try:
            with AutoClean(lambda: DB.session.rollback()):
                DB.session.add(user)
                DB.session.flush()
                user.homeserverID, user.maildir = Servers.allocUser(user.ID, props.get("homeserver"))
                with UserSetup(user, DB.session) as us:
                    us.run()
                if chat:
                    try:
                        user.chat = chat
                    except ValueError as err:
                        logger.error("Failed to activate chat: "+err.args[0])
                DB.session.commit()
                if not us.success:
                    return "Error during user setup: "+us.error, us.errorCode
                if sync:
                    try:
                        user.syncStore()
                    except Exception:
                        pass
                    return user, 201
        except IntegrityError as err:
            return "Object violates database constraints "+err.orig.args[1], 400
        except Exception as err:
            DB.session.rollback()
            return "Failed to create user "+" - ".join(str(arg) for arg in err.args), 500

    @classmethod
    def mkContact(cls, props, externID=None, *args, **kwargs):
        smtpaddress = props.get("username")
        props["status"] = Users.CONTACT
        props["username"] = "temp"
        error = Users.checkCreateParams(props)
        if smtpaddress:
            props["properties"][PropTags.SMTPADDRESS] = smtpaddress
        props["properties"][PropTags.DISPLAYTYPEEX] = Users.REMOTE_MAILUSER
        if error is not None:
            return error, 400
        try:
            user = Users(props)
            user.externID = externID
        except (InvalidAttributeError, MismatchROError, MissingRequiredAttributeError, ValueError) as err:
            return err.args[0], 400
        try:
            DB.session.add(user)
            DB.session.flush()
            user.username = "contact-"+str(user.ID)
            DB.session.commit()
            return user, 201
        except IntegrityError as err:
            return "Object violates database constraints "+err.orig.args[1], 400
        except Exception as err:
            DB.session.rollback()
            return "Failed to create contact"+" - ".join(str(arg) for arg in err.args), 500

    @classmethod
    def _commit(*args, **kwargs):
        with Service("systemd", errors=Service.SUPPRESS_ALL) as sysd:
            sysd.reloadService("gromox-http.service", "gromox-zcore.service")

    @validates("username")
    def validateUsername(self, key, value, *args):
        if not self.username == value and Users.query.filter(Users.altname == value).count():
            raise ValueError("Username is already used as alternative name")
        return value

    @validates("altname")
    def validateAltname(self, key, value, *args):
        if Users.query.filter(Users.username == value).count():
            raise ValueError("Alternative name is already in use")
        return value or None

    def syncStore(self):
        """Write all properties to the exmdb store.

        For users without a store, this function has no effect.
        """
        if not self.maildir:
            return
        with Service("exmdb") as exmdb:
            props = []
            for tag, value in self.properties.rawmap().items():
                try:
                    props.append(exmdb.TaggedPropval(tag, value))
                except Exception:
                    pass
            client = exmdb.user(self)
            client.setStoreProperties(0, props)


class UserProperties(DB.Base):
    __tablename__ = "user_properties"

    supportedTypes = PropTypes.intTypes | PropTypes.floatTypes | {PropTypes.STRING, PropTypes.WSTRING}

    userID = Column("user_id", INTEGER(unsigned=True), ForeignKey(Users.ID, ondelete="cascade", onupdate="cascade"), primary_key=True)
    tag = Column("proptag", INTEGER(unsigned=True), primary_key=True)
    orderID = Column("order_id", INTEGER(unsigned=True), primary_key=True, server_default="1")
    _propvalbin = Column("propval_bin", VARBINARY(4096))
    _propvalstr = Column("propval_str", VARCHAR(4096))

    user = relationship(Users, back_populates="_properties")

    def __init__(self, tag, value, user):
        if tag & 0x0FFF not in self.supportedTypes:
            raise NotImplementedError("Prop type is currently not supported")
        self.tag = tag
        self.val = value
        self.user = user

    @property
    def name(self):
        if self.tag is None:
            return None
        tagname = PropTags.lookup(self.tag, None)
        return tagname.lower() if tagname else "<unknown>"

    @name.setter
    def name(self, value):
        tag = getattr(PropTags, value.upper(), None)
        if tag is None:
            raise ValueError("Unknown PropTag '{}'".format(value))
        if tag & 0x0FFF not in self.supportedTypes:
            raise ValueError("{}: Tag type {} is not supported".format(PropTags.lookup(tag), PropTypes.lookup(tag)))
        self.tag = tag

    @property
    def type(self):
        return self.tag & 0xFFFF

    @property
    def baseType(self):
        return self.tag & 0x0FFF

    @property
    def val(self):
        if self.type == PropTypes.BINARY:
            return self._propvalbin
        if self.type == PropTypes.FILETIME:
            return datetime.fromtimestamp(nxTime(int(self._propvalstr))).strftime("%Y-%m-%d %H:%M:%S")
        return PropTypes.pyType(self.type)(self._propvalstr)

    @val.setter
    def val(self, value):
        value = PropTags.convertValue(self.tag, value)
        if self.type == PropTypes.BINARY:
            self._propvalbin = value
        else:
            self._propvalstr = str(value)

    @property
    def content(self):
        return self._propvalbin if self.baseType == PropTypes.BINARY else self._propvalstr


class Aliases(DataModel, DB.Base, NotifyTable):
    __tablename__ = "aliases"

    aliasname = Column("aliasname", VARCHAR(128), nullable=False, unique=True, primary_key=True)
    mainname = Column("mainname", VARCHAR(128), ForeignKey(Users.username, ondelete="cascade", onupdate="cascade"),
                      nullable=False, index=True)

    main = relationship(Users, back_populates="aliases")

    _dictmapping_ = ((Text("aliasname", flags="init"),),
                     (Text("mainname", flags="init"),))

    def __init__(self, aliasname, main, *args, **kwargs):
        if main.ID == 0:
            raise ValueError("Cannot alias superuser")
        self.main = main
        self.fromdict(aliasname)

    def fromdict(self, aliasname, *args, **kwargs):
        if not formats.email.match(aliasname):
            raise ValueError("'{}' is not a valid email address".format(aliasname))
        self.aliasname = aliasname
        return self

    @classmethod
    def _commit(*args, **kwargs):
        with Service("systemd", errors=Service.SUPPRESS_ALL) as sysd:
            sysd.reloadService("gromox-delivery.service", "gromox-http.service", "gromox-zcore.service")


class Fetchmail(DataModel, DB.Base):
    __tablename__ = "fetchmail"

    _sa = ("password", "kerberos_v5", "kerberos", "kerberos_v4", "gssapi", "cram-md5", "otp", "ntlm", "msn", "ssh", "any")

    ID = Column("id", INTEGER(10, unsigned=True), nullable=False, primary_key=True)
    userID = Column("user_id", INTEGER(10, unsigned=True), ForeignKey(Users.ID), nullable=False)
    mailbox = Column("mailbox", VARCHAR(255), nullable=False)
    active = Column("active", TINYINT(1, unsigned=True), nullable=False, server_default="1")
    srcServer = Column("src_server", VARCHAR(255), nullable=False)
    srcAuth = Column("src_auth", ENUM(*_sa), nullable=False, server_default="password")
    srcUser = Column("src_user", VARCHAR(255), nullable=False)
    srcPassword = Column("src_password", VARCHAR(255), nullable=False)
    srcFolder = Column("src_folder", VARCHAR(255), nullable=False, default="")
    fetchall = Column("fetchall", TINYINT(1, unsigned=True), nullable=False, server_default="0")
    keep = Column("keep", TINYINT(1, unsigned=True), nullable=False, server_default="1")
    protocol = Column("protocol", ENUM("POP3", "IMAP", "POP2", "ETRN", "AUTO"), nullable=False, server_default='IMAP')
    useSSL = Column("usessl", TINYINT(1, unsigned=True), nullable=False, server_default="1")
    sslCertCheck = Column("sslcertck", TINYINT(1, unsigned=True), nullable=False, server_default="0")
    sslCertPath = Column("sslcertpath", VARCHAR(255, charset="utf8"))
    sslFingerprint = Column("sslfingerprint", VARCHAR(255, charset="latin1"))
    extraOptions = Column("extra_options", TEXT)
    date = Column("date", TIMESTAMP, nullable=False, server_default="current_timestamp()", onupdate=func.current_timestamp())

    user = relationship(Users)

    _dictmapping_ = ((Id(), Text("mailbox", flags="patch"),
                      Bool("active", flags="patch"),
                      Text("srcServer", flags="patch"),
                      Text("srcUser", flags="patch"),
                      Date("date", time=True),
                      Text("srcAuth", flags="patch"),
                      Text("srcPassword", flags="patch"),
                      Text("srcFolder", flags="patch"),
                      Bool("fetchall", flags="patch"),
                      Bool("keep", flags="patch"),
                      Text("protocol", flags="patch"),
                      Bool("useSSL", flags="patch"),
                      Bool("sslCertCheck", flags="patch"),
                      Text("sslCertPath", flags="patch"),
                      Text("sslFingerprint", flags="patch"),
                      Text("extraOptions", flags="patch")),
                     (Id("userID", flags="hidden"),))

    def __init__(self, props, user, *args, **kwargs):
        self.user = user
        if "mailbox" not in props:
            self.mailbox = user.username
        self.fromdict(props)

    @validates("mailbox")
    def validateMailbox(self, key, value, *args):
        if not formats.email.match(value):
            raise ValueError("'{}' is not a valid e-mail address".format(value))
        return value

    @validates("srcPassword")
    def validatePassword(self, key, value, *args):
        return value.replace("'", "").replace('"', '')

    @validates("srcServer")
    def validateServer(self, key, value, *args):
        if not formats.domain.match(value.split(":", 1)[0]):
            raise ValueError("'{}' is not a valid domain".format(value))
        return value

    def __str__(self):
        fetchoptions = "options"
        if self.useSSL == 1:
            fetchoptions += " ssl"
            if self.sslFingerprint:
                fetchoptions = " sslfingerprint '{}'".format(self.sslFingerprint)
            if self.sslCertCheck == 1:
                fetchoptions += " sslCertCheck"
            if self.sslCertPath:
                fetchoptions += "sslcertpath "+self.sslCertPath
        if self.fetchall == 0:
            fetchoptions += " no"
        fetchoptions += " fetchall"
        fetchoptions += " keep" if self.keep == 1 else " nokeep"
        if self.extraOptions:
            fetchoptions += " "+self.extraOptions
        server, service = self.srcServer.split(":", 1) if ":" in self.srcServer else (self.srcServer, None)
        service = " service "+service if service else ""
        srcFolder = " folder "+self.srcFolder if self.srcFolder and self.protocol not in ("POP3", "ETRN", "ODMR") else ""
        return "poll {} with proto {}{} user {} there with password '{}' is {} here{} {}\n"\
            .format(server, self.protocol, service, self.srcUser, self.srcPassword, self.mailbox, srcFolder, fetchoptions)


# Available as of n93
class UserDevices(DataModel, DB.Base):
    __tablename__ = "user_devices"

    ID = Column("id", INTEGER(unsigned=True), primary_key=True)
    userID = Column("user_id", INTEGER(unsigned=True), ForeignKey(Users.ID), nullable=False)
    deviceID = Column("device_id", VARCHAR(64), nullable=False)
    status = Column("status", INTEGER(unsigned=True), nullable=False, server_default="0")

    STATUS_NA = 0
    STATUS_OK = 1 << 0
    STATUS_PENDING = 1 << 1
    STATUS_REQUESTED = 1 << 2
    STATUS_WIPED = 1 << 3

    DEFAULT = {"status": STATUS_NA}

    _dictmapping_ = ((Id(), Id("userID", flags="init"), Id("deviceID", flags="init"), Int("status", flags="patch")),)


class UserDeviceHistory(DataModel, DB.Base):
    __tablename__ = "user_device_history"

    ID = Column("id", INTEGER(unsigned=True), primary_key=True)
    userDeviceID = Column("user_device_id", INTEGER(unsigned=True), ForeignKey(UserDevices.ID), nullable=False)
    time = Column("time", TIMESTAMP, nullable=False, server_default="now()")
    remoteIP = Column("remote_ip", VARCHAR(64), nullable=True)
    status = Column("status", INTEGER(unsigned=True), nullable=False, server_default="0")

    _dictmapping_ = ((Id(), Id("userDeviceID", flags="init")),
                     (DataModel.Prop("time", func=lambda date: int(date.timestamp()) if date else None, flags="init,sort"),
                      Text("remoteIP", flags="init"),
                      Int("status", flags="init")))


class UserSecondaryStores(DB.Base):
    __tablename__ = "secondary_store_hints"

    primaryID = Column("primary", INTEGER(unsigned=True), ForeignKey(Users.ID), primary_key=True)
    secondaryID = Column("secondary", INTEGER(unsigned=True), ForeignKey(Users.ID), primary_key=True)


class Forwards(DataModel, DB.Base):
    __tablename__ = "forwards"

    ID = Column("id", INTEGER(10, unsigned=True), nullable=False, primary_key=True)
    username = Column("username", VARCHAR(128), ForeignKey(Users.username), nullable=False, unique=True)
    forwardType = Column("forward_type", TINYINT, nullable=False, server_default="0")
    destination = Column("destination", VARCHAR(128), nullable=False)

    user = relationship(Users)

    _dictmapping_ = ((Text("forwardType", flags="patch"),
                      Text("destination", flags="patch")),
                     (Id(), Text("username", flags="patch")))

    CC = 0
    REDIRECT = 1

    def __init__(self, props, user, *args, **kwargs):
        self.user = user
        self.fromdict(props)

    @validates("destination")
    def validateDestination(self, key, value, *args):
        if not formats.email.match(value):
            raise ValueError("'{}' is not a valid e-mail address".format(value))
        return value


from .domains import Domains
from . import misc, mlists, roles


Users.NTregister()
Aliases.NTregister()

if sqlalchemy.__version__.split(".") >= ["1", "4"]:
    inspect(Users).add_property("orgID", column_property(select(Domains.orgID)
                                                         .where(Domains.ID == Users.domainID).scalar_subquery()))
else:
    inspect(Users).add_property("orgID", column_property(select([Domains.orgID])
                                                         .where(Domains.ID == Users.domainID).as_scalar()))


@event.listens_for(Users, "expire")
def _User_expire(target, *args, **kwargs):
    if target is not None:
        target._propcache = None



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, InvalidUseError
from .common import proptagCompleter

from argparse import ArgumentParser

_statusMap = {0: "active", 1: "suspended", 2: "out-of-date", 3: "deleted", 4: "shared", 5: "contact"}
_statusColor = {0: "green", 1: "yellow", 2: "yellow", 3: "red", 4: "cyan", 5: "blue"}
_userAttributes = ("ID", "aliases", "changePassword", "chat", "chatAdmin", "domainID", "forward", "homeserverID", "lang",
                   "ldapID", "maildir", "pop3_imap", "privArchive", "privChat", "privFiles", "privVideo", "publicAddress",
                   "smtp", "status", "username")
_deviceStatus = {0: "unknown", 1: "ok", 2: "pending", 4: "requested", 8: "wiped", 16: "pending (account)",
                 32: "requested (account)", 64: "wiped (account)"}
_deviceStatusStyle = {0: {"attrs": ["dark"]},
                      1: {"color": "green"},
                      2: {"color": "yellow"},
                      4: {"color": "red"},
                      8: {"color": "light_red"},
                      16: {"color": "yellow"},
                      32: {"color": "red"},
                      64: {"color": "light_red"}}


def _mkUserQuery(args):
    from .common import userFilter
    from orm.users import Users
    query = Users.query
    if "userspec" in args:
        query = query.filter(userFilter(args.userspec))
    if "filter" in args and args.filter is not None:
        query = Users.autofilter(query, {f.split("=", 1)[0]: f.split("=", 1)[1] for f in args.filter if "=" in f})
    if "sort" in args and args.sort is not None:
        query = Users.autosort(query, args.sort)
    return query


def _mkStatus(cli, status):
    return cli.col(_statusMap.get(status, "unknown"), _statusColor.get(status, "magenta"))


def _dumpUser(cli, user, indent=0):
    def privstr():
        from orm.users import Users
        bits = user.privilegeBits
        privs = []
        if bits & Users.USER_PRIVILEGE_POP3_IMAP:
            privs.append("pop_imap")
        if bits & Users.USER_PRIVILEGE_SMTP:
            privs.append("smtp")
        if bits & Users.USER_PRIVILEGE_CHGPASSWD:
            privs.append("passwd")
        if bits & Users.USER_PRIVILEGE_PUBADDR:
            privs.append("pubaddr")
        if bits & Users.USER_PRIVILEGE_CHAT:
            privs.append("chat")
        if bits & Users.USER_PRIVILEGE_VIDEO:
            privs.append("video")
        if bits & Users.USER_PRIVILEGE_FILES:
            privs.append("files")
        if bits & Users.USER_PRIVILEGE_ARCHIVE:
            privs.append("archive")
        if len(privs) == 0:
            return ""
        return "("+",".join(privs)+")"

    from ldap3.utils.conv import escape_filter_chars
    homeserver = cli.col("(local)", attrs=["dark"]) if user.homeserver is None else \
        "{} ({})".format(user.homeserver.ID, user.homeserver.hostname)
    cli.print("{}ID: {}".format(" "*indent, user.ID))
    cli.print("{}username: {}".format(" "*indent, user.username))
    cli.print("{}domainID: {}".format(" "*indent, user.domainID))
    cli.print("{}homeserver: {}".format(" "*indent, homeserver))
    cli.print("{}lang: {}".format(" "*indent, user.lang or cli.col("(not set)", attrs=["dark"])))
    cli.print("{}maildir: {}".format(" "*indent, user.maildir or cli.col("(not set)", attrs=["dark"])))
    cli.print("{}privilegeBits: {} {}".format(" "*indent, user.privilegeBits, cli.col(privstr(), attrs=["dark"])))
    cli.print("{}addressStatus: {} ({}|{})".format(" "*indent, user.addressStatus,
                                                   _mkStatus(cli, user.domainStatus), _mkStatus(cli, user.status)))
    cli.print(" "*indent+"ldapID: "+(escape_filter_chars(user.externID) if user.externID is not None else
                                       cli.col("(none)", attrs=["dark"])))
    cli.print(" "*indent+"chat: "+(user.chatID if user.chatID else cli.col("(none)", attrs=["dark"])) +
              (" ("+cli.col("inactive", "red")+")" if user.chatID and not user.chat else ""))
    if user.chat:
        cli.print(" "*indent+"chatAdmin: "+(cli.col("yes", "yellow") if user.chatAdmin else "no"))
    cli.print(" "*indent+"aliases:"+(cli.col(" (none)", attrs=["dark"]) if len(user.aliases) == 0 else ""))
    for alias in user.aliases:
        cli.print(" "*indent+"  "+alias.aliasname)
    cli.print(" "*indent+"roles:"+(cli.col(" (none)", attrs=["dark"]) if len(user.roles) == 0 else ""))
    for role in user.roles:
        cli.print(" "*indent+"  "+role.name)
    cli.print(" "*indent+"fetchmail:"+(cli.col(" (none)", attrs=["dark"]) if len(user.fetchmail) == 0 else ""))
    for fml in user.fetchmail:
        cli.print("{}  {}@{}/{} ({})".format(" "*indent, fml.srcUser, fml.srcServer, fml.srcFolder,
                                             cli.col("active", "green") if fml.active == 1 else cli.col("inactive", "red")))
    cli.print(" "*indent+"properties:"+(cli.col(" (none)", attrs=["dark"]) if len(user.properties) == 0 else ""))
    for key, value in user.properties.items():
        cli.print("{}  {}: {}".format(" "*indent, key, value))


def _getUser(args, requireMailbox=False):
    cli = args._cli
    users = _mkUserQuery(args).all()
    if len(users) == 0:
        cli.print(cli.col("No user matching '{}'.".format(args.userspec), "yellow"))
        return 1, None
    if len(users) > 1:
        cli.print("'{}' is ambiguous. Candidates are:".format(args.userspec))
        for user in users:
            cli.print("  {}:\t{}".format(user.ID, user.username))
        return 2, None
    if requireMailbox and not users[0].maildir:
        cli.print(cli.col(f"User '{users[0].username}' has no mailbox", "red"))
        return 3, None
    return 0, users[0]


def _splitData(args):
    cliargs = {"_handle", "_cli", "userspec", "no_defaults"}
    data = {}
    attributes = data["attributes"] = {key: value for key, value in args.items() if value is not None and key not in cliargs}
    data["aliases"] = attributes.pop("alias", None) or ()
    data["aliases_rm"] = attributes.pop("remove_alias", None) or ()
    data["props"] = attributes.pop("property", None) or []
    data["props_rm"] = attributes.pop("remove_property", None) or []
    data["storeprops"] = attributes.pop("storeprop", None) or []
    data["storeprops_rm"] = attributes.pop("remove_storeprop", None) or []
    data["noldap"] = attributes.pop("no_ldap", False)
    data["delchat"] = attributes.pop("delete_chat_user", False)
    return data


def _updateStoreprops(cli, user, props, props_rm=()):
    if not (props or props_rm):
        return
    from tools.constants import PropTags, ExchangeErrors
    import pyexmdb
    add, remove = [], []
    for pv in props:
        prop, val = pv.split("=", 1)
        try:
            prop, val = PropTags.normalize(prop, val)
            add.append(pyexmdb.TaggedPropval(prop, val))
        except ValueError as err:
            cli.print(cli.col("Failed to set store property '{}': {}".format(prop, err.args[0]), "yellow"))
        except TypeError:
            cli.print(cli.col("Failed to set store property '{}': Tag type not supported"), "yellow")
    for p in props_rm:
        try:
            prop = PropTags.deriveTag(p)
            remove.append(prop)
        except ValueError as err:
            cli.print(cli.col("Failed to set store property '{}': {}".format(prop, err.args[0]), "yellow"))
    if not (add or remove):
        return
    from services import Service, ServiceUnavailableError
    try:
        with Service("exmdb") as exmdb:
            client = exmdb.user(user)
            if len(remove):
                client.removeStoreProperties(remove)
            if len(add):
                problems = client.setStoreProperties(0, add)
                if problems:
                    problems = [(PropTags.lookup(entry.proptag, hex(entry.proptag)).lower(),
                                 ExchangeErrors.lookup(entry.err, hex(entry.err)))
                                for entry in problems]
                    cli.print(cli.col("Problems where encountered setting the following tags:\n\t" +
                                      "\n\t".join("{} ({})".format(tag, err) for tag, err in problems), "yellow"))
    except (pyexmdb.ExmdbError, ServiceUnavailableError) as err:
        cli.print(cli.col("Failed to set store properties: "+err.args[0], "yellow"))


def cliUserShow(args):
    cli = args._cli
    cli.require("DB")
    users = _mkUserQuery(args).all()
    if len(users) == 0:
        cli.print(cli.col("No users found.", "yellow"))
        return 1
    for user in users:
        cli.print(cli.col("{} ({}):".format(user.username, user.ID), attrs=["bold"]))
        _dumpUser(cli, user, 2)


def cliUserList(args):
    cli = args._cli
    cli.require("DB")
    cli.print(cli.col("The 'user list' command is deprecated and may be removed in the future. Use 'user query' instead.\n",
                      "yellow"))
    users = _mkUserQuery(args).all()
    if len(users) == 0:
        cli.print(cli.col("No users found.", "yellow"))
        return 1
    maxNameLen = max(len(user.username) for user in users)
    for user in users:
        if user.domainName() is not None:
            printName = "{}@{}".format(cli.col(user.baseName(), attrs=["bold"]), user.domainName())
        else:
            printName = cli.col(user.username, attrs=["bold"])
        cli.print("{}:\t{}{}({}|{})".format(user.ID, printName, " "*(maxNameLen-len(user.username)+4),
                                            _mkStatus(cli, user.domainStatus), _mkStatus(cli, user.status)))
    cli.print("({} users total)".format(len(users)))


def cliUserCreate(args):
    cli = args._cli
    cli.require("DB")
    from orm.domains import Domains
    from orm.misc import DBConf
    from orm.users import DB, Users
    from tools.misc import RecursiveDict
    if args.no_defaults:
        props = {}
    else:
        props = DBConf.getFile("grommunio-admin", "defaults-system", True).get("user", RecursiveDict())
        if "@" in args.username:
            domain = Domains.query.filter(Domains.domainname == args.username.split("@", 1)[1]).with_entities(Domains.ID).first()
            if domain is not None:
                props.update(DBConf.getFile("grommunio-admin", "defaults-domain-"+str(domain.ID), True).get("user", {}))
    data = _splitData(args.__dict__)
    props.update(data["attributes"])
    props["username"] = args.username
    props["aliases"] = data["aliases"]
    properties = data["properties"] = {}
    if args.domain:
        from .common import domainCandidates
        domain = domainCandidates(args.domain).with_entities(Domains.ID).all()
        if len(domain) == 0:
            cli.print(cli.col("Domain not found.", "red"))
            return 2
        if len(domain) != 1:
            cli.print(cli.col("Domain specification is ambiguous.", "red"))
            return 3
        props["domainID"] = domain[0].ID

    for pv in data["props"]+data["storeprops"]:
        if "=" in pv:
            prop, val = pv.split("=", 1)
            properties[prop] = val
    props["properties"] = properties
    result, code = Users.mkContact(props) if props.get("status", 0) == Users.CONTACT else Users.create(props)
    if code != 201:
        cli.print(cli.col("Could not create user: "+result, "red"))
        return 1
    DB.session.commit()
    _updateStoreprops(cli, result, data["storeprops"])
    _dumpUser(cli, result)


def cliUserDelete(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from tools.misc import GenericObject
    ret, user = _getUser(args)
    if ret:
        return ret
    userdata = GenericObject(maildir=user.maildir, homeserver=user.homeserver)
    if not args.yes:
        if cli.confirm("Delete user '{}' ({})? [y/N]: ".format(user.username, user.ID)) != Cli.SUCCESS:
            return 3
    else:
        cli.print("Deleting user '{}' ({})".format(user.username, user.ID))
    user.delete(not args.keep_chat)
    DB.session.commit()
    cli.print("User deleted.")
    if userdata.maildir == "":
        cli.print("No user files to delete.")
        return 0
    cli.print("Unloading store...", end="", flush=True)

    from services import Service
    with Service("exmdb", errors=Service.SUPPRESS_INOP) as exmdb:
        client = exmdb.user(user)
        client.unloadStore()
        cli.print("Done.", end="")
    cli.print("")
    if args.keep_files or (not args.yes and cli.confirm("Delete user directory from disk? [y/N]: ") != Cli.SUCCESS):
        cli.print(cli.col("Files remain in "+userdata.maildir, attrs=["bold"]))
        return 0
    cli.print("Deleting user files...", end="")
    import shutil
    shutil.rmtree(userdata.maildir, ignore_errors=True)
    cli.print("Done.")


def _cliUserDevicesDecodeSyncState(args, data, username):
    import base64
    import json
    try:
        data = json.loads(base64.b64decode(data))["data"]
        if "devices" in data:
            data = data["devices"][username]["data"]
        return data
    except Exception as err:
        cli = args._cli
        cli.print(cli.col("Failed to decode sync state ({}: {}"
                          .format(type(err).__name__, " - ".join(str(arg) for arg in err.args))))
        return {}


def _cliUserDevicesGetDevices(args, user):
    from services import Service
    from tools.config import Config
    from orm.users import UserDevices
    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        syncStates = client.getSyncData(Config["sync"]["syncStateFolder"])
    requested = args.device if "device" in args else ()
    syncStates = {device: _cliUserDevicesDecodeSyncState(args, state, user.username) for device, state in syncStates.items()
                  if not requested or device in requested}
    wipeStatus = {device.deviceID: device.status for device in UserDevices.query.filter(UserDevices.userID == user.ID)
                  if not requested or device.deviceID in requested}
    for device, state in syncStates.items():
        state["wipeStatus"] = wipeStatus.pop(device, 0)
    for device, status in wipeStatus.items():
        syncStates[device] = {"wipeStatus": status}
    return syncStates


def _mkDate(cli, timestamp):
    if not timestamp:
        return ""
    from datetime import datetime
    now = datetime.now()
    date = datetime.fromtimestamp(timestamp)
    datestr = date.strftime("%H:%M")
    if date.day-now.day or date.month-now.month or date.year-now.year:
        datestr += cli.col(date.strftime(" %d.%m.%Y"), attrs=["dark"])
    return datestr


def _mkWipeStatus(cli, status):
    return cli.col(_deviceStatus.get(status, f"<{status}>"), **_deviceStatusStyle.get(status, {"color": "magenta"}))


def cliUserDevicesList(args):
    from .common import Table

    cli = args._cli
    ret, user = _getUser(args, requireMailbox=True)
    if ret:
        return ret

    data = _cliUserDevicesGetDevices(args, user)
    devices = [(cli.col(device, attrs=["bold"]), state.get("devicefriendlyname", ""), state.get("useragent", ""),
                str(state.get("asversion", "")), _mkDate(cli, state.get("lastupdatetime")),
                _mkWipeStatus(cli, state["wipeStatus"]))
               for device, state in data.items()]
    devices = sorted(devices, key=lambda entry: entry[0])
    Table(devices, ("ID", "Device", "Agent", "Version", "Updated", "Status"), empty=cli.col("(No devices)", attrs=["dark"]))\
        .print(cli)


def cliUserDevicesShow(args):
    from .common import Table
    cli = args._cli
    ret, user = _getUser(args, requireMailbox=True)
    if ret:
        return ret

    tf = {"firstsynctime": _mkDate, "lastupdatetime": _mkDate, "wipeStatus": _mkWipeStatus}
    keys = ("devicetype", "devicemodel", "deviceos", "useragent", "devicemobileoperator", "deviceimei", "deviceoslanguage",
            "deviceuser", "firstsynctime", "lastupdatetime", "asversion", "announcedasversion", "hierarchyuuid", "wipeStatus")
    devices = _cliUserDevicesGetDevices(args, user)
    for device, state in devices.items():
        cli.print(cli.col(device, attrs=["bold"]))
        data = [("  "+key+":", str(tf.get(key, lambda _, x: x)(cli, state[key]))) for key in keys if key in state]
        Table(data).print(cli)


def cliUserDevicesRemoveResync(args):
    from orm.users import DB, UserDevices
    from services import Service
    from tools.config import Config

    cli = args._cli
    ret, user = _getUser(args, requireMailbox=True)
    if ret:
        return ret

    with Service("exmdb") as exmdb:
        client = exmdb.user(user)
        if args.action == "remove" and not args.device:
            client.removeSyncStates(Config["sync"]["syncStateFolder"])
            cli.print(cli.col("Removed all devices", "green"))
            UserDevices.query.filter(UserDevices.userID == user.ID).delete()
            DB.session.commit()
            return

        devices = args.device or [device for device, state in _cliUserDevicesGetDevices(args, user).items()
                                  if "deviceid" in state]
        for device in devices:
            if args.action == "remove":
                client.removeDevice(Config["sync"]["syncStateFolder"], device)
                UserDevices.query.filter(UserDevices.userID == user.ID, UserDevices.deviceID == device).delete()
                cli.print(f"Removed {device}")
            else:
                client.resyncDevice(Config["sync"]["syncStateFolder"], device, user.ID)
                cli.print(f"Removed states of {device}")


def cliUserDeviceWipe(args):
    from datetime import datetime
    from orm.users import DB, UserDevices, UserDeviceHistory

    cli = args._cli
    ret, user = _getUser(args, requireMailbox=True)
    if ret:
        return ret

    device = UserDevices.query.filter(UserDevices.userID == user.ID, UserDevices.deviceID == args.device).first()
    currentStatus = device.status if device is not None else 0
    requestedStatus = 1 if args.mode == "cancel" else 16 if args.mode == "account" else 2
    if requestedStatus == currentStatus:
        cli.print("Nothing to do.")
        return
    if currentStatus in (4, 8, 32, 64):
        cli.print("Cannot modify while wipe already in progress")
        return 10
    if device is None:
        device = UserDevices(dict(userID=user.ID, deviceID=args.device, status=requestedStatus))
        DB.session.add(device)
        DB.session.flush()
    else:
        device.status = requestedStatus
    DB.session.add(UserDeviceHistory(dict(userDeviceID=device.ID, time=datetime.utcnow(), remoteIP="[CLI]",
                                          status=device.status)))
    DB.session.commit()
    if requestedStatus == 1:
        cli.print(cli.col(f"Cancelled wipe on {args.device}"))
    else:
        cli.print(cli.col(f"Requested {args.mode} wipe for device {args.device}", on_color="on_red", attrs=["bold"]))


def cliUserModify(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from orm.users import Aliases
    ret, user = _getUser(args)
    if ret:
        return ret
    data = _splitData(args.__dict__)
    if data["noldap"]:
        user.externID = None
    if data["delchat"]:
        from services import Service
        with Service("chat") as chat:
            chat.deleteUser(user)
    try:
        user.fromdict(data["attributes"])
        if data["aliases"]:
            existing = {a.aliasname for a in user.aliases}
            [Aliases(alias, user) for alias in data["aliases"] if alias not in existing]
        if data["aliases_rm"]:
            user.aliases = [alias for alias in user.aliases if alias.aliasname not in data["aliases_rm"]]
        for pv in data["props"]+data["storeprops"]:
            try:
                prop, val = pv.split("=", 1)
                user.properties[prop] = val
            except (KeyError, ValueError) as err:
                cli.print(cli.col("Failed to set property '{}': {}".format(prop, err.args[0]), "yellow"))
        for prop in data["props_rm"]+data["storeprops_rm"]:
            try:
                user.properties[prop] = None
            except KeyError:
                cli.print(cli.col("Failed to remove property '{}'".format(prop), "yellow"))
    except ValueError as err:
        cli.print(cli.col("Failed to update user: "+err.args[0], "red"))
        return 1
    DB.session.commit()
    _updateStoreprops(cli, user, data["storeprops"], data["storeprops_rm"])
    _dumpUser(cli, user)


def cliUserQuery(args):
    def mkUsername(value):
        if "@" not in value:
            return cli.col(value, attrs=["bold"])
        username, domain = value.split("@", 1)
        return cli.col(username, attrs=["bold"])+"@"+domain

    cli = args._cli
    cli.require("DB")

    attrTf = {"username": mkUsername,
              "status": lambda v: cli.col(str(v)+"/", attrs=["dark"])+_mkStatus(cli, v)}\
        if args.format == "pretty" else {}

    from .common import Table
    from orm.users import Users
    args.attributes = args.attributes or ("ID", "username", "status")
    query = _mkUserQuery(args)
    query = Users.optimize_query(query, args.attributes)
    users = [user.todict(args.attributes) for user in query]
    separator = args.separator or ("," if args.format == "csv" else "  ")
    data = [[attrTf.get(attr, lambda x: x)(user.get(attr)) for attr in args.attributes] for user in users]
    header = None if len(args.attributes) <= 1 and args.format == "pretty" else args.attributes
    table = Table(data, header, separator, cli.col("(no results)", attrs=["dark"]))
    table.dump(cli, args.format)


def _cliUserspecCompleter(prefix, **kwargs):
    from orm.users import Users
    return (user.username for user in Users.query.filter(Users.username.ilike(prefix+"%"))
                                                 .with_entities(Users.username).all())


def _cliParseStatus(value):
    try:
        return int(value)
    except Exception:
        pass
    from orm.users import Users
    value = value.lower()
    if value == "normal":
        return Users.NORMAL
    if value == "suspended":
        return Users.SUSPENDED
    if value in ("outofdate", "out-of-date"):
        return Users.OUTOFDATE
    if value == "deleted":
        return Users.DELETED
    if value == "shared":
        return Users.SHARED
    if value == "contact":
        return Users.CONTACT
    raise ValueError("Unknown user status '{}'".format(value))


def _cliAddUserAttributes(parser: ArgumentParser):
    def getBool(val):
        if not val.isdigit() and val.lower() not in ("yes", "no", "true", "false", "y", "n"):
            raise ValueError("'{}' cannot be interpreted as boolean.".format(val))
        return bool(int(val)) if val.isdigit() else val in ("yes", "true", "y")

    def optBool(val):
        if not val.isdigit() and val.lower() not in ("yes", "no", "true", "false", "y", "n"):
            return val
        return bool(int(val)) if val.isdigit() else val in ("yes", "true", "y")

    def assignment(arg):
        if "=" not in arg:
            raise ValueError("'{}' is not an assignment".format(arg))
        return arg

    def proptagAssignCompleter(*args, **kwargs):
        return proptagCompleter(*args, **kwargs, addSuffix="=")

    parser.add_argument("--changePassword", type=getBool, metavar="<bool>", help="Whether the user can change the password")
    parser.add_argument("--chat", type=optBool, metavar="<bool>", help="Whether to create a chat user")
    parser.add_argument("--chatAdmin", type=getBool, metavar="<bool>", help="Whether the user has chat admin privileges")
    parser.add_argument("--domain", metavar="DOMAINSPEC", help="Domain name or ID to create the user for")
    parser.add_argument("--homeserver", type=int, metavar="ID", help="ID of the home server or 0 for local user")
    parser.add_argument("--lang", help="User store language")
    parser.add_argument("--ldapID", help="Identifier of the LDAP object linked to the user")
    parser.add_argument("--pop3-imap", type=getBool, metavar="<bool>", help="Whether the user has the POP3/IMAP privilege")
    parser.add_argument("--privArchive", type=getBool, metavar="<bool>", help="Whether the user has the archiving privilege")
    parser.add_argument("--privChat", type=getBool, metavar="<bool>", help="Whether the user has the chat privilege")
    parser.add_argument("--privFiles", type=getBool, metavar="<bool>", help="Whether the user has the files privilege")
    parser.add_argument("--privVideo", type=getBool, metavar="<bool>", help="Whether the user has the video privilege")
    parser.add_argument("--public-address", type=getBool, metavar="<bool>",
                        help="Whether the user has the public address privilege")
    parser.add_argument("--smtp", type=getBool, metavar="<bool>", help="Whether the user has the SMTP privilege")
    parser.add_argument("--status", type=_cliParseStatus, help="User address status")

    parser.add_argument("--alias", action="append", help="Add alias")
    parser.add_argument("--property", action="append", type=assignment, metavar="propspec=value",
                        help="Set property defined by propspec to value").completer = proptagAssignCompleter
    parser.add_argument("--storeprop", action="append", type=assignment, metavar="propspec=value",
                        help="Set store property defined by propspec to value").completer = proptagAssignCompleter


def _setupCliUser(subp: ArgumentParser):
    class AttrChoice:
        def __contains__(self, value):
            return value == [] or value in _userAttributes

        def __getitem__(self, i):
            return _userAttributes[i]

        def __len__(self):
            return len(_userAttributes)

    def deviceParser(parent, action, handler, **kwargs):
        sub = parent.add_parser(action, **kwargs)
        sub.set_defaults(action=action, _handle=handler)
        sub.add_argument("device", nargs="*", help="Device ID")
        return sub

    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    create = sub.add_parser("create",  help="Create user")
    create.add_argument("username", help="E-Mail address of the user")
    create.add_argument("--no-defaults", action="store_true", help="Do not apply configured default values")
    create.set_defaults(_handle=cliUserCreate)
    _cliAddUserAttributes(create)
    delete = sub.add_parser("delete", help="Delete user")
    delete.set_defaults(_handle=cliUserDelete)
    delete.add_argument("userspec", help="User ID or name").completer = _cliUserspecCompleter
    delete.add_argument("-c", "--keep-chat", action="store_true", help="Do not permanently delete the chat user")
    delete.add_argument("-k", "--keep-files", action="store_true", help="Do not delete files on disk")
    delete.add_argument("-y", "--yes", action="store_true", help="Do not ask for confirmation")
    device = sub.add_parser("devices", help="User device management")
    device.add_argument("userspec", help="User ID or name").completer = _cliUserspecCompleter
    deviceActions = device.add_subparsers()
    deviceParser(deviceActions, "list", cliUserDevicesList, help="List devices of a user")
    deviceParser(deviceActions, "show", cliUserDevicesShow, help="Show detailed information about a device")
    deviceParser(deviceActions, "remove", cliUserDevicesRemoveResync, help="Show detailed information about a device")
    deviceParser(deviceActions, "resync", cliUserDevicesRemoveResync, help="Show detailed information about a device")
    deviceWipe = deviceActions.add_parser("wipe")
    deviceWipe.set_defaults(_handle=cliUserDeviceWipe)
    deviceWipe.add_argument("device", help="Device ID")
    deviceWipe.add_argument("--mode", default="normal", choices=("account", "cancel", "normal"), help="Set device wipe mode")
    list = sub.add_parser("list", help="List users")
    list.set_defaults(_handle=cliUserList)
    list.add_argument("userspec", nargs="?", help="User ID or name prefix")
    list.add_argument("-f", "--filter", action="append", help="Filter by attribute, e.g. -f ID=42")
    list.add_argument("-s", "--sort", action="append", help="Sort by attribute, e.g. -s username,desc")
    modify = sub.add_parser("modify",  help="Modify user")
    modify.set_defaults(_handle=cliUserModify)
    modify.add_argument("userspec", help="User ID or name prefix").completer = _cliUserspecCompleter
    _cliAddUserAttributes(modify)
    modify.add_argument("--delete-chat-user", action="store_true", help="Permanently delete chat user")
    modify.add_argument("--no-ldap", action="store_true", help="Unlink user from ldap object")
    modify.add_argument("--remove-alias", metavar="ALIAS", action="append", help="Remove alias")
    modify.add_argument("--remove-property", action="append", metavar="propspec", help="Remove property from user")
    modify.add_argument("--remove-storeprop", action="append", metavar="propspec", help="Remove property from user's store")
    modify.add_argument("--username", help="Rename user")
    query = sub.add_parser("query", help="Query specific user attributes")
    query.set_defaults(_handle=cliUserQuery)
    query.add_argument("-f", "--filter", action="append", help="Filter by attribute, e.g. -f ID=42")
    query.add_argument("--format", choices=("csv", "json-flat", "json-structured", "pretty"), help="Set output format",
                       metavar="FORMAT", default="pretty")
    query.add_argument("--separator", help="Set column separator")
    query.add_argument("-s", "--sort", action="append", help="Sort by attribute, e.g. -s username,desc")
    query.add_argument("attributes", nargs="*", choices=AttrChoice(), help="Attributes to query", metavar="ATTRIBUTE")
    show = sub.add_parser("show", help="Show detailed information about user")
    show.set_defaults(_handle=cliUserShow)
    show.add_argument("userspec", help="User ID or name").completer = _cliUserspecCompleter
    show.add_argument("-f", "--filter", action="append", help="Filter by attribute, e.g. -f ID=42")
    show.add_argument("-s", "--sort", action="append", help="Sort by attribute, e.g. -s username,desc")


@Cli.command("user", _setupCliUser, help="User management")
def cliUserStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, ArgumentParser, InvalidUseError


class CTrace:
    UNUSED = 0
    USED = 1
    EXTENDED = 2
    SHADOWED = 3
    NOCHANGE = 4

    KEY = 0
    LIST = 1
    DICT = 2

    mark = ("x ", "+ ", "* ", "~ ")
    _style = ({"color": "red"},
              {"color": "green"},
              {"color": "yellow"},
              {"color": "white", "attrs": ["dark"]},
              {"color": "red", "attrs": ["dark"]})

    _fbStyle = {"attrs": ["dark"]}

    def __init__(self, value, filename=None):
        value = {} if value is None else value  # Interpret empty config files as empty dict instead of None
        self.state = self.USED
        self.type = self.LIST if isinstance(value, list) else self.DICT if isinstance(value, dict) else self.KEY
        self.value = {key: CTrace(val, filename) for key, val in value.items()} if self.type == self.DICT else\
                     [(val, filename) for val in value] if self.type == self.LIST else value
        self.overwritten = []
        self.overwrites = False
        self.filename = filename

    def __contains__(self, key):
        """Check if trace contains a key (dot-notation supported)."""
        if key is None:
            return True
        if self.type != self.DICT:
            return False
        if "." in key:
            pre, post = key.split(".", 1)
            return pre in self.value and post in self.value[pre]
        return key in self.value

    def __getitem__(self, key):
        """Get trace by key (dot-notation supported)."""
        if key is None:
            return self
        if self.type != self.DICT:
            raise KeyError(key)
        if "." in key:
            pre, post = key.split(".", 1)
            return self.value[pre][post]
        return self.value[key]

    @property
    def style(self):
        """Generate style for file tracing."""
        st = dict(self._style[self.state])
        if self.overwrites:
            st["attrs"] = st.get("attrs", [])+["bold"]
        return st

    def fstyle(self, stylemap):
        """Generate style for key-tracing."""
        return {"attrs": ["bold"]} if self.overwritten and self.type in (self.DICT, self.LIST) else \
               stylemap.get(self.filename, self._fbStyle) if stylemap else self.style

    def update(self, file, entry):
        """Add future update to trace"""
        newstate = self.UNUSED if self.state == self.UNUSED or self.type != entry.type else\
                   self.EXTENDED if self.type in (self.LIST, self.DICT) and self.type == entry.type else\
                   self.SHADOWED if self.value == entry.value else self.UNUSED
        if self.type == self.DICT and self.state != self.UNUSED:
            for key, value in self.value.items():
                if key in entry.value:
                    value.update(file, entry.value[key])
        self.overwritten.append((file, self.NOCHANGE if self.state == newstate == self.UNUSED else newstate))
        self.state = newstate
        entry.overwrites = True

    def merge(self, entry):
        """Merge another trace into self."""
        file = entry.filename
        oldstate = self.EXTENDED if self.type in (self.LIST, self.DICT) and self.type == entry.type else\
                   self.SHADOWED if self.value == entry.value else self.UNUSED
        self.overwritten.append((self.filename or "default", oldstate))
        self.filename = file
        if self.type == entry.type and self.type in (self.DICT, self.LIST):
            if self.type == self.DICT:
                for key, value in entry.value.items():
                    if key in self.value:
                        self.value[key].merge(value)
                    else:
                        self.value[key] = value
            elif self.type == self.LIST:
                self.value += entry.value
        else:
            self.value = entry.value

    def printval(self, args, hist=0, indent=0, stylemap=None, prefix=0):
        """Print value of a trace."""
        cli = args._cli
        if hist == 0:
            ovstr = ""
        elif stylemap is None:
            ovstr = "["+", ".join(cli.col(ov[0], **self._style[ov[1]]) for ov in self.overwritten)+"]" \
                    if self.state != self.USED else ""
        elif args.show_history or self.type in (self.DICT, self.LIST):
            sfile = cli.col(self.filename, **stylemap.get(self.filename, self._fbStyle))
            if self.overwritten and self.type == self.KEY:
                sfile = cli.col(sfile, attrs=["underline"])
            ovstr = "["+", ".join(cli.col(ov[0], **stylemap.get(ov[0], self._fbStyle)) for ov in self.overwritten) +\
                    (", " if self.overwritten else "")+sfile+"]"
        else:
            ovstr = "["+cli.col(self.filename, **stylemap.get(self.filename, self._fbStyle))+"]"
        if self.type == self.LIST and len(self.value):
            lines = []
            for value, filename in self.value:
                style = stylemap.get(filename, self._fbStyle) if stylemap else\
                        {"color": "red" if self.state == self.UNUSED else "green"}
                data = "\n"+("" if stylemap else self.mark[self.state])+" "*(indent+3)+"- "
                ov = "["+cli.col(filename, **stylemap.get(filename, self._fbStyle))+"]" if stylemap else ""
                if isinstance(value, (dict, list)):
                    import yaml
                    mark = "" if stylemap else self.mark[self.state]
                    data = ["\n"+mark+" "*(indent+5)+rep for rep in yaml.dump(value).split("\n") if rep]
                    lines += [cli.col(line, **style)+" "*(hist-len(line)+1)+ov for line in data]
                else:
                    vstr = str(value)
                    lines.append(cli.col(data+str(value), **style)+(" "*(hist-len(data)-len(vstr)+1)+ov if stylemap else ""))
            return " "*(hist-prefix)+ovstr+"".join(lines)
        res = str(self.value) if self.type == self.KEY else "" if len(self.value) else "{}"
        post = " "*(hist-len(res)-prefix)+ovstr if hist and (self.state != self.USED or stylemap) else ""
        return cli.col(res, **self.fstyle(stylemap))+post

    def printFile(self, args, hist=0, indent=0):
        """Print trace in by-file style."""
        cli = args._cli
        if self.type == self.DICT:
            for key, value in self.value.items():
                base = self.mark[value.state]+" "*(indent+1)+key+": "
                cli.print(cli.col(base, **value.style), end="")
                cli.print(value.printval(args, hist, indent, prefix=len(base)))
                value.printFile(args, hist, indent+2)
        elif indent == 0:
            cli.print(self.printval(args, hist))

    def printConfig(self, args, hist, stylemap, indent=0):
        """Print trace in by-key style."""
        cli = args._cli
        if self.type == self.DICT:
            for key, value in self.value.items():
                base = " "*(indent)+key+": "
                cli.print(cli.col(base, **value.fstyle(stylemap)), end="")
                cli.print(value.printval(args, hist, indent, stylemap, len(base)))
                value.printConfig(args, hist, stylemap, indent+2)
        elif indent == 0:
            cli.print(self.printval(args, hist, 0, stylemap))

    def inlwidth(self):
        """Return inline value print width."""
        return 0 if self.type in (self.LIST, self.DICT) and len(self.value) else len(str(self.value))+1

    def maxwidth(self, indent=0):
        """Return maximum line width when printing."""
        if self.type == self.DICT:
            if len(self.value):
                return max(max(len(key)+1+value.inlwidth() for key, value in self.value.items())+indent,
                           max(value.maxwidth(indent+2) for value in self.value.values()))
        elif self.type == self.LIST and len(self.value):
            import yaml
            return max(len(self.mark[self.state]+" "*(indent+2)+rep) for rep in yaml.dump(self.value).split("\n") if rep)
        return len(str(self.value)) if indent == 0 else 0


def cliConfigCheck(args):
    from tools.config import validate
    cli = args._cli
    result = validate()
    if result is None:
        cli.print("Configuration schema valid")
        return 0
    else:
        cli.print(cli.col(result, "red"))
        return 1


def cliConfigGet(args):
    from tools.config import Config
    from .common import getKey, NotFound
    keyspec = getattr(args, "key", "")
    c = getKey(Config, keyspec.split("."))
    if isinstance(c, NotFound):
        rep = "\n"
    elif isinstance(c, (dict, list)):
        import yaml
        rep = yaml.dump(c)
    else:
        rep = str(c)
    args._cli.print(rep)


def _traceFiles(args):
    import yaml
    from os import scandir
    from tools.config import _defaultConfig
    cli = args._cli
    defaultConfig = _defaultConfig()
    files = [("default", CTrace(defaultConfig))]
    try:
        with open("config.yaml", encoding="utf-8") as file:
            fconf = yaml.load(file, Loader=yaml.SafeLoader)
    except Exception as err:
        cli.print(cli.col("Failed to open main config file: "+" - ".join(str(arg) for arg in err.args), "yellow"))
        return 1
    upd = CTrace(fconf)
    files[0][1].update("./config.yaml", upd)
    files.append(("./config.yaml", upd))
    confdir = fconf.get("confdir")
    if not confdir:
        return files, confdir
    configFiles = sorted([file.path for file in scandir(confdir) if file.name.endswith(".yaml")])
    for configFile in configFiles:
        try:
            with open(configFile, encoding="utf-8") as file:
                fconf = yaml.load(file, Loader=yaml.SafeLoader)
            configFile = configFile.replace(confdir, "$CONFD")
            upd = CTrace(fconf)
            for old in files:
                old[1].update(configFile, upd)
            files.append((configFile, upd))
        except Exception as err:
            cli.print(cli.col("Failed to open '{}' config file: {}"
                              .format(configFile, " - ".join(str(arg) for arg in err.args)), "yellow"))
    return files, confdir


def _traceKeys(args):
    import yaml
    from os import scandir
    from tools.config import _defaultConfig
    cli = args._cli
    config = CTrace(_defaultConfig(), "default")
    try:
        with open("config.yaml", encoding="utf-8") as file:
            fconf = yaml.load(file, Loader=yaml.SafeLoader)
    except Exception as err:
        cli.print(cli.col("Failed to open main config file: "+" - ".join(str(arg) for arg in err.args), "yellow"))
        return 1
    config.merge(CTrace(fconf, "./config.yaml"))
    confdir = fconf.get("confdir")
    if not confdir:
        return config, []
    configFiles = sorted([file.path for file in scandir(confdir) if file.name.endswith(".yaml")])
    for configFile in configFiles:
        try:
            with open(configFile, encoding="utf-8") as file:
                fconf = yaml.load(file, Loader=yaml.SafeLoader)
            configFile = configFile.replace(confdir, "$CONFD")
            upd = CTrace(fconf, configFile)
            config.merge(upd)
        except Exception as err:
            cli.print(cli.col("Failed to open '{}' config file: {}"
                              .format(configFile, " - ".join(str(arg) for arg in err.args)), "yellow"))
    return config, [configFile.replace(confdir, "$CONFD") for configFile in configFiles]


def _printByFile(args):
    cli = args._cli
    files, confdir = _traceFiles(args)
    files = [f for f in files if args.key in f[1]]
    if len(files) == 0:
        cli.print(cli.col("No matching files", "yellow"))
        return
    if any(f[0].startswith("$CONFD") for f in files):
        cli.print("Note: $CONFD evaluates to '{}'\n".format(cli.col(confdir, attrs=["underline"])))
    hist = max(f[1][args.key].maxwidth() for f in files if args.key in f[1])+4 if args.show_history and len(files) else 0
    for file in files:
        cli.print("<"+cli.col(file[0], attrs=["bold"])+">")
        file[1][args.key].printFile(args, hist)
        cli.print()


def _printByKey(args):
    cli = args._cli
    from itertools import cycle, product
    config, files = _traceKeys(args)
    hist = config[args.key].maxwidth()+4 if args.key in config else 0
    colors = ("yellow", "green", "red", "cyan", "magenta")
    attrs = [[("dark", "bold", "underline")[i] for i in (0, 1, 2) if index & (1 << i)] for index in range(8)]
    styles = {filename: {"color": style[1], "attrs": style[0]}
              for filename, style in zip(files, cycle(product(attrs, colors)))}
    styles["default"] = styles["./config.yaml"] = {}
    if len(files):
        cli.print("Configuration resulting from " +
                  ", ".join(cli.col(file, **styles.get(file, {"attrs": ["dark"]})) for file in files)+".\n" +
                  "$CONFD evaluates to '{}'".format(cli.col(config["confdir"].value, attrs=["underline"]))+".\n")
    if args.key in config:
        config[args.key].printConfig(args, hist, styles)
    else:
        cli.print(cli.col("'{}' not found.".format(args.key), "yellow"))


def cliConfigTrace(args):
    return (_printByFile if args.mode == "files" else _printByKey)(args)


def _configKeyspecCompleter(prefix, **kwargs):
    from tools.config import Config
    from .common import getKey
    split = prefix.split(".")
    path, prefix = split[:-1], split[-1]
    parent = getKey(Config, path)
    if isinstance(parent, dict):
        path = ".".join(path)+"." if path else ""
        return list(path+key for key in parent.keys())
    return ()


def _setupCliConfigParser(subp: ArgumentParser):
    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    check = sub.add_parser("check", help="Check if configuration is valid")
    check.set_defaults(_handle=cliConfigCheck)
    get = sub.add_parser("get", help="Print configuration values to stdout", aliases=["dump"])
    get.set_defaults(_handle=cliConfigGet)
    get.add_argument("key", nargs="?", default="", help="Configuration key to print").completer = _configKeyspecCompleter
    trace = sub.add_parser("trace", help="Analyse where effective configuration values are set")
    trace.set_defaults(_handle=cliConfigTrace)
    trace.add_argument("mode", choices=("files", "values"), help="Display trace by file or by key")
    trace.add_argument("key", nargs="?", help="Only show trace for specific key").completer = _configKeyspecCompleter
    trace.add_argument("-s", "--show-history", action="store_true", help="Show which files overwrite a value")


@Cli.command("config", _setupCliConfigParser, help="Show or check configuration")
def cliConfigStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from . import Cli

from argparse import ArgumentParser

import random
import string
from datetime import datetime
from getpass import getpass

passwordChars = string.ascii_letters+string.digits+'!"#$%&()*+-/;<=>?[]_{|}~'
defaultPassLength = 16


def mkPasswd(length=None):
    """Generate random password.

    Parameters
    ----------
    length : int, optional
        Length of the password. The default is `defaultPassLength` (16).

    Returns
    -------
    str
        String containing the password.
    """
    return "".join(random.choices(passwordChars, k=length or defaultPassLength))


def createAdmin():
    """Create admin user.

    Returns
    -------
    admin : Users
        The admin user object
    adminPass : str
        Clear text password
    """
    from orm.users import Users
    adminPass = mkPasswd()
    admin = Users(None)
    admin.ID = 0
    admin.username = "admin"
    admin.realName = "System Administrator"
    admin.domainID = 0
    admin.password = adminPass
    admin.maxSize = 0
    admin.createDay = datetime.now()
    return admin, adminPass


def _passwdParserSetup(subp: ArgumentParser):
    subp.add_argument("user", nargs="?", action="store", type=str,
                      help="User to change the password of. If omitted, set password of system administrator.")
    subp.add_argument("--auto", "-a", action="store_true", help="Automatically generate password.")
    subp.add_argument("--length", "-l", action="store", type=int, default=defaultPassLength,
                      help="Length of auto-generated password (default {})".format(defaultPassLength))
    subp.add_argument("--password", "-p", action="store", type=str,
                      help="New password. If neither -p nor -a are specified, the new password is set interactively.")


@Cli.command("passwd", _passwdParserSetup, help="User password management")
def setUserPassword(args):
    cli = args._cli
    cli.require("DB")
    from orm.users import DB, Users
    import orm.roles
    if args.user is not None:
        from .common import userCandidates
        users = userCandidates(args.user).all()
        if len(users) == 0:
            cli.print(cli.col("User '{}' not found.".format(args.user), "yellow"))
            return 1
        if len(users) != 1:
            cli.print(cli.col("'{}' is ambiguous.".format(args.user), "yellow"))
            return 1
        user = users[0]
        if user.externID is not None:
            cli.print(cli.col("Cannot change password of LDAP user", "yellow"))
            return 2
    else:
        user = Users.query.filter(Users.ID == 0).first()
        if user is None:
            cli.print("System admin user not found, creating...")
            DB.session.execute("SET sql_mode='NO_AUTO_VALUE_ON_ZERO';")
            user, _ = createAdmin()
            DB.session.add(user)
    cli.print("Setting password for user '{}'".format(user.username))
    if args.auto:
        password = mkPasswd(args.length)
        cli.print("New password is "+cli.col(password, attrs=["bold"]))
    elif args.password is not None:
        password = args.password
    else:
        try:
            password = cli.input("Password: ", secret=True)
            if cli.input("Retype password: ", secret=True) != password:
                cli.print(cli.col("Passwords do not match, aborting.", "yellow"))
                return 3
        except (KeyboardInterrupt, EOFError):
            cli.print(cli.col("\nAborted.", "yellow"))
            return 4
    user.password = password
    DB.session.commit()
    cli.print("Password updated")



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from .. import defaultListHandler

import api
from api.core import API, secure
from api.security import checkPermissions

from tools.permissions import SystemAdminPermission, DomainAdminROPermission, OrgAdminPermission
from tools.dnsHealth import fullDNSCheck

from flask import request, jsonify
from sqlalchemy import or_


@API.route(api.BaseRoute+"/domains", methods=["GET"])
@secure(requireDB=True, authLevel="user")
def getAvailableDomains():
    from orm.domains import Domains
    permissions = request.auth["user"].permissions()
    if SystemAdminPermission() in permissions:
        domainFilters = ()
    else:
        domainIDs = {permission.domainID for permission in permissions if isinstance(permission, DomainAdminROPermission)}
        orgIDs = {permission.orgID for permission in permissions if isinstance(permission, OrgAdminPermission)}
        domainFilters = () if "*" in domainIDs or "*" in orgIDs else \
                        (or_(Domains.ID.in_(domainIDs), Domains.orgID.in_(orgIDs)),)
    return defaultListHandler(Domains, filters=domainFilters)


@API.route(api.BaseRoute+"/domains/<int:domainID>/dnsCheck", methods=["GET"])
@secure(requireDB=True)
def checkDomainDNS(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).with_entities(Domains.domainname).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    dnsCheck = fullDNSCheck(domain.domainname)
    return jsonify(dnsCheck)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH
"""
Module containing admin Managed CONFigurations
"""

from .config import Config
from .misc import setDirectoryOwner, setDirectoryPermission
from services import Service

import logging
logger = logging.getLogger("mconf")

LDAP = {}
AUTHMGR = {}

censorship = {
    "LDAP": {"ldap_bind_pass", "ldap_bind_user"}
}

_ldapDepServices = ("gromox-http.service", "gromox-midb.service", "gromox-zcore.service", "gromox-delivery.service",
                    "gromox-delivery-queue.service", "gromox-imap.service", "gromox-pop3.service")


def _loadConf(path):
    from multidict import MultiDict
    try:
        with open(path, encoding="utf-8") as file:
            conf = MultiDict()
            for line in file:
                if line.strip().startswith("#") or "=" not in line:
                    continue
                key, value = line.split("=", 1)
                conf.add(key.strip(), value.strip())
        return conf
    except FileNotFoundError:
        logger.info("Could not load '{}': file not found".format(path))
        return MultiDict()


def _fDumpConf(file, conf, censor=()):
    file.write("# Configuration automatically generated by grommunio-admin.\n")
    for key, value in conf.items():
        if isinstance(value, list) and key not in censor:
            file.writelines(("{}={}\n".format(key, entry) for entry in value))
        else:
            file.write("{}={}\n".format(key, "***" if key in censor else value))


def _dumpConf(path, conf):
    with open(path, "w", encoding="utf-8") as file:
        _fDumpConf(file, conf)
    uid = Config["mconf"].get("fileUid", Config["options"].get("fileUid"))
    gid = Config["mconf"].get("fileGid", Config["options"].get("fileGid"))
    prm = Config["mconf"].get("filePermissions", Config["options"].get("filePermissions"))
    setDirectoryOwner(path, uid, gid)
    setDirectoryPermission(path, prm)


def _addIfDef(dc, d, sc, s, all=False, type=None):
    def tf(v):
        return v if type is None else type(v)
    if s in sc:
        dc[d] = tf(sc[s]) if not all else [tf(v) for v in sc.getall(s)]


def _reloadServices(*services):
    with Service("systemd", errors=Service.SUPPRESS_ALL) as sysd:
        sysd.tryReloadRestartService(*services)

###############################################################################


def _transformLdap(conf):
    LDAP = {"connection": {}, "groups": {}, "users": {}}
    _addIfDef(LDAP, "disabled", conf, "ldap_disabled", type=lambda x: x.lower() in ("true", "yes", "1"))
    _addIfDef(LDAP["connection"], "server", conf, "ldap_host")
    _addIfDef(LDAP["connection"], "bindUser", conf, "ldap_bind_user")
    _addIfDef(LDAP["connection"], "bindPass", conf, "ldap_bind_pass")
    _addIfDef(LDAP["connection"], "starttls", conf, "ldap_start_tls", type=lambda x: x.lower() in ("true", "yes", "1"))
    _addIfDef(LDAP["connection"], "connections", conf, "data_connections", type=int)
    _addIfDef(LDAP, "baseDn", conf, "ldap_search_base")
    _addIfDef(LDAP, "objectID", conf, "ldap_object_id")
    _addIfDef(LDAP["users"], "username", conf, "ldap_mail_attr")
    _addIfDef(LDAP["users"], "filters", conf, "ldap_user_filters", all=True)
    _addIfDef(LDAP["users"], "filter", conf, "ldap_user_filter")
    _addIfDef(LDAP["users"], "contactFilter", conf, "ldap_contact_filter")
    _addIfDef(LDAP["users"], "searchAttributes", conf, "ldap_user_search_attrs", all=True)
    _addIfDef(LDAP["users"], "displayName", conf, "ldap_user_displayname")
    _addIfDef(LDAP["users"], "defaultQuota", conf, "ldap_user_default_quota", type=int)
    _addIfDef(LDAP["users"], "templates", conf, "ldap_user_templates", all=True)
    _addIfDef(LDAP["users"], "aliases", conf, "ldap_user_aliases")
    _addIfDef(LDAP["groups"], "groupaddr", conf, "ldap_group_addr")
    _addIfDef(LDAP["groups"], "groupfilter", conf, "ldap_group_filter")
    _addIfDef(LDAP["groups"], "groupname", conf, "ldap_group_name")
    _addIfDef(LDAP["groups"], "groupMemberAttr", conf, "ldap_group_memberof")
    if "ldap_user_attributes" in conf:
        LDAP["users"]["attributes"] = {entry.split(" ", 1)[0]: entry.split(" ", 1)[1]
                                       for entry in conf.getall("ldap_user_attributes") if " " in entry}
    return LDAP


def _flattenLdap(conf):
    LDAP = {}
    _addIfDef(LDAP, "ldap_disabled", conf, "disabled")
    if "connection" in conf:
        _addIfDef(LDAP, "ldap_host", conf["connection"], "server")
        _addIfDef(LDAP, "ldap_bind_user", conf["connection"], "bindUser")
        _addIfDef(LDAP, "ldap_bind_pass", conf["connection"], "bindPass")
        _addIfDef(LDAP, "ldap_start_tls", conf["connection"], "starttls")
    _addIfDef(LDAP, "ldap_search_base", conf, "baseDn")
    _addIfDef(LDAP, "ldap_object_id", conf, "objectID")
    if "users" in conf:
        _addIfDef(LDAP, "ldap_mail_attr", conf["users"], "username")
        _addIfDef(LDAP, "ldap_user_displayname", conf["users"], "displayName")
        _addIfDef(LDAP, "ldap_user_filters", conf["users"], "filters")
        _addIfDef(LDAP, "ldap_user_filter", conf["users"], "filter")
        _addIfDef(LDAP, "ldap_contact_filter", conf["users"], "contactFilter")
        _addIfDef(LDAP, "ldap_user_search_attrs", conf["users"], "searchAttributes")
        _addIfDef(LDAP, "ldap_user_default_quota", conf["users"], "defaultQuota")
        _addIfDef(LDAP, "ldap_user_templates", conf["users"], "templates")
        _addIfDef(LDAP, "ldap_user_aliases", conf["users"], "aliases")
        if "attributes" in conf["users"]:
            LDAP["ldap_user_attributes"] = ["{} {}".format(key, value) for key, value in conf["users"]["attributes"].items()]
    if "groups" in conf:
        _addIfDef(LDAP, "ldap_group_addr", conf["groups"], "groupaddr")
        _addIfDef(LDAP, "ldap_group_filter", conf["groups"], "groupfilter")
        _addIfDef(LDAP, "ldap_group_name", conf["groups"], "groupname")
        _addIfDef(LDAP, "ldap_group_memberof", conf["groups"], "groupMemberAttr")
    return LDAP


def loadLdap():
    """(Re)load LDAP configuration from disk.

    Note that this function only populates the configuration but does not validate or deploy it to the ldap module.

    Returns
    -------
    str
        Error message or None if successful
    """
    if "ldapPath" not in Config["mconf"]:
        return "mconf.ldapPath not set"
    try:
        global LDAP
        LDAP = _transformLdap(_loadConf(Config["mconf"]["ldapPath"]))
    except Exception as err:
        return " - ".join((str(arg) for arg in err.args))


def dumpLdap(conf=None, file=None, reloadServices=True, censor=False):
    """Write LDAP configuration to disk.

    Parameters
    ----------
    conf : dict, optional
        New LDAP configuration or None to use current config.

    Returns
    -------
    str
        Error message or None if successful
    """
    if "ldapPath" not in Config["mconf"]:
        return "mconf.ldapPath not set"
    try:
        global LDAP
        if conf is not None:
            LDAP = conf
        if file is None:
            _dumpConf(Config["mconf"]["ldapPath"], _flattenLdap(LDAP))
        else:
            _fDumpConf(file, _flattenLdap(LDAP), censorship["LDAP"] if censor else ())
        if reloadServices:
            _reloadServices(*_ldapDepServices)
    except Exception as err:
        return " - ".join((str(arg) for arg in err.args))


###############################################################################


def loadAuthmgr():
    """(Re)load authmgr configuration from disk.

    Returns
    -------
    str
        Error message or None if successful
    """
    if "authmgrPath" not in Config["mconf"]:
        return "mconf.authmgrPath not set"
    try:
        global AUTHMGR
        rconf = _loadConf(Config["mconf"]["authmgrPath"])
        AUTHMGR = {"authBackendSelection": rconf.get("auth_backend_selection", "always_mysql")}
    except Exception as err:
        return " - ".join((str(arg) for arg in err.args))


def dumpAuthmgr(conf=None, file=None, reloadServices=False):
    """Write authmgr configuration to disk.

    Parameters
    ----------
    conf : dict, optional
        New authmgr configuration or None to use current config.

    Returns
    -------
    str
        Error message or None if successful
    """
    if "authmgrPath" not in Config["mconf"]:
        return "mconf.authmgrPath not set"
    try:
        global AUTHMGR
        if conf is not None:
            AUTHMGR = conf
        wconf = {"auth_backend_selection": AUTHMGR.get("authBackendSelection", "always_mysql")}
        if file is None:
            _dumpConf(Config["mconf"]["authmgrPath"], wconf)
            if reloadServices:
                _reloadServices(*_ldapDepServices)
        else:
            _fDumpConf(file, wconf)
    except Exception as err:
        return " - ".join((str(arg) for arg in err.args))


###############################################################################


def load():
    error = loadLdap()
    if error:
        logger.error("Could not load ldap config: "+error)
    error = loadAuthmgr()
    if error:
        logger.error("Could not load authmgr config: "+error)


load()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH


from argparse import ArgumentParser

from . import Cli, InvalidUseError

_configs = ("authmgr", "ldap")


def _getConfig(cli, name):
    if name not in _configs:
        cli.print(cli.col("Unknown config '{}'".format(name), "red"))
        return None
    from tools import mconf
    if name == "authmgr":
        return mconf.AUTHMGR
    elif name == "ldap":
        return mconf.LDAP


def _cliMconfPrint(args):
    cli = args._cli
    config = _getConfig(cli, args.config)
    if config is None:
        return 1
    import yaml
    data = yaml.dump(config)
    cli.print(data, end="" if data.endswith("\n") else "\n")


def _cliMconfDump(args):
    cli = args._cli
    from tools import mconf
    if args.config == "authmgr":
        mconf.dumpAuthmgr(file=cli.stdout)
    elif args.config == "ldap":
        mconf.dumpLdap(file=cli.stdout, censor=args.censor)
    else:
        cli = cli.print(cli.col("Invalid config", "red"))
        return 1


def _getValue(args):
    if args.int:
        return int(args.value, 0)
    elif args.bool:
        if args.value.lower() not in ("y", "n", "yes", "no", "true", "false", "1", "0"):
            raise ValueError("invalid boolean value '{}'".format(args.value))
        return args.value.lower() in ("y", "yes", "true", "1")
    else:
        return args.value


def _cliMconfReload(args):
    cli = args._cli
    from tools import mconf
    if args.config == "authmgr":
        error = mconf.loadAuthmgr()
    elif args.config == "ldap":
        error = mconf.loadLdap()
    else:
        cli.print(cli.col("Unknown config '{}'".format(args.config), "red"))
        return 1
    if error:
        cli.print(cli.col("Could not load {} config: {}".format(args.config, error), "yellow"))
        return 2


def _cliMconfSave(args):
    cli = args._cli
    from tools import mconf
    try:
        if args.config == "authmgr":
            error = mconf.dumpAuthmgr(file=cli.open("mconf.authmgrPath", "w", cli.fs is None))
        elif args.config == "ldap":
            error = mconf.dumpLdap(file=cli.open("mconf.ldapPath", "w", cli.fs is None))
        else:
            cli.print(cli.col("Unknown config '{}'".format(args.config), "red"))
            return 1
    except KeyError as err:
        error = err.args[0]
    except Exception as err:
        error = " - ".join((str(arg) for arg in err.args))
    cli.print("Configuration saved" if error is None else cli.col("Failed to save configuration: "+error, "yellow"))


def _cliMconfModify(args):
    if "action" not in args or args.action not in ("add", "remove", "set", "unset"):
        args.parser.print_usage()
        return -2
    cli = args._cli
    config = _getConfig(cli, args.config)
    if config is None:
        return 1
    if "." in args.key:
        path, var = args.key.rsplit(".", 1)
        parent = config
        for level in path.split("."):
            if level not in parent:
                parent[level] = {}
            elif not isinstance(parent[level], dict):
                cli.print(cli.col("'{}': invalid path".format(args.key), "red"))
                return 2
            parent = parent[level]
    else:
        parent, var = config, args.key
    if args.action in ("set", "add", "remove"):
        try:
            value = _getValue(args)
        except ValueError as err:
            cli.print(cli.col(err.args[0], "red"))
            return 3
    if args.action == "set":
        parent[var] = value
    elif args.action == "unset":
        parent.pop(var, None)
    elif args.action == "add":
        target = parent.get(var)
        if target is None:
            parent[var] = target = []
        elif not isinstance(target, list):
            cli.print(cli.col("Cannot add value: '{}' is not an array".format(var), "red"))
            return 4
        target.append(value)
    elif args.action == "remove":
        target = parent.get(var)
        if not isinstance(target, list):
            cli.print(cli.col("Cannot remove value: '{}' is not an array".format(var), "red"))
            return 4
        try:
            target.remove(value)
        except ValueError:
            cli.print(cli.col("Value {} not found in '{}'".format(value, var), "red"))
            return 5
    if not args.defer:
        _cliMconfSave(args)


def _mconfKeyCompleter(prefix, parsed_args, **kwargs):
    from .common import getKey
    config = _getConfig(None, parsed_args.config)
    if config is None:
        return ()
    split = prefix.split(".")
    path, prefix = split[:-1], split[-1]
    parent = getKey(config, path)
    if isinstance(parent, dict):
        path = ".".join(path)+"." if path else ""
        return list(path+key for key in parent.keys()
                    if parsed_args.action not in ("add", "remove") or isinstance(parent[key], (list, dict)))
    return ()


def _setupCliMconfAddValue(parser: ArgumentParser):
    typearg = parser.add_mutually_exclusive_group()
    typearg.add_argument("-b", "--bool", action="store_true", help="Cast value to boolean")
    typearg.add_argument("-i", "--int", action="store_true", help="Cast value to integer")
    parser.add_argument("value")


def _setupCliMconf(subp: ArgumentParser):
    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    dump = sub.add_parser("dump", help="Dump configuration file to stdout")
    dump.set_defaults(_handle=_cliMconfDump)
    dump.add_argument("config", choices=_configs)
    dump.add_argument("-c", "--censor", action="store_true", help="Hide confidential information")
    modify = sub.add_parser("modify", help="Modify configuration")
    modify.set_defaults(_handle=_cliMconfModify, parser=modify)
    modify.add_argument("config", choices=_configs)
    modify.add_argument("-d", "--defer", action="store_true", help="Do not write changes to disk")
    modifysub = modify.add_subparsers()
    modifyadd = modifysub.add_parser("add", help="Add value to configuration list")
    modifyadd.set_defaults(action="add")
    modifyadd.add_argument("key").completer = _mconfKeyCompleter
    _setupCliMconfAddValue(modifyadd)
    modifyremove = modifysub.add_parser("remove", help="Remove value from configuration list")
    modifyremove.set_defaults(action="remove")
    modifyremove.add_argument("key").completer = _mconfKeyCompleter
    _setupCliMconfAddValue(modifyremove)
    modifyset = modifysub.add_parser("set", help="Set configuration value")
    modifyset.set_defaults(action="set")
    modifyset.add_argument("key").completer = _mconfKeyCompleter
    _setupCliMconfAddValue(modifyset)
    modifyunset = modifysub.add_parser("unset", help="Unset configuration value")
    modifyunset.set_defaults(action="unset")
    modifyunset.add_argument("key").completer = _mconfKeyCompleter
    printConf = sub.add_parser("print", help="Show configuration")
    printConf.set_defaults(_handle=_cliMconfPrint)
    printConf.help = "Print current configuration"
    printConf.add_argument("config", choices=_configs)
    reload = sub.add_parser("reload", help="Reload configuration")
    reload.set_defaults(_handle=_cliMconfReload)
    reload.help = "Reload configuration from disk"
    reload.add_argument("config", choices=_configs)
    save = sub.add_parser("save", help="Write configuration to disk")
    save.set_defaults(_handle=_cliMconfSave)
    save.add_argument("config", choices=_configs)


@Cli.command("mconf", _setupCliMconf, help="Managed configurations manipulation")
def cliMconfStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, ArgumentParser, InvalidUseError

_states = {-1: ("UNLOADED", {"attrs": ["dark"]}),
           0: ("LOADED", {"color": "green"}),
           1: ("UNAVAILABLE", {"color": "yellow"}),
           2: ("SUSPENDED", {"color": "yellow"}),
           3: ("ERROR", {"color": "red"}),
           4: ("DISABLED", {"color": "red", "attrs": ["dark"]})}


def _counts(cli, service):
    from services import ServiceHub
    if service.state == ServiceHub.UNAVAILABLE:
        return cli.col(" ({}/{})".format(service.failures, service.maxfailures or "\u221e"), **_sstyle(service))
    elif service.state == ServiceHub.SUSPENDED:
        return cli.col(" ({}/{})".format(service.reloads, service.maxreloads), **_sstyle(service))
    elif service.state == ServiceHub.ERROR:
        if service.maxreloads:
            return cli.col(" ({}/{})".format(service.reloads, service.maxreloads), **_sstyle(service))
        elif service.maxfailures:
            return cli.col(" ({}/{})".format(service.failures, service.maxfailures or "\u221e", **_sstyle(service)))
    return ""


def _excinfo(service, verbose):
    if not all((verbose, service.exc, service.state >= 1)):
        return ""
    if verbose > 1:
        return " ("+" - ".join(str(arg) for arg in service.exc.args)+")"
    return " ("+(str(service.exc.args[0]) if len(service.exc.args) else type(service.exc).__name__)+")"


def _sname(service):
    return _states.get(service.state, ("UNKNOWN",))[0]


def _sstyle(service):
    return _states.get(service.state, (None, {"color": "magenta"}))[1]


def _printService(cli, service):
    cli.print(cli.col(service.name, attrs=["bold"])+"  "+cli.col(_sname(service), **_sstyle(service)) +
              _counts(cli, service)+_excinfo(service, 2))


def cliServiceLoad(args):
    cli = args._cli
    from services import ServiceHub
    try:
        service = ServiceHub.load(args.service, *args.args, force_reload=args.reload)
        _printService(cli, service)
    except ValueError as err:
        cli.print(cli.col("Failed to load service: "+err.args[0], "red"))


def cliServiceStatus(args):
    from .common import Table
    from services import ServiceHub
    cli = args._cli

    data = []
    services = args.services or ServiceHub.services()
    for service in services:
        instances = ServiceHub.instances(service)
        data.append((cli.col(service, attrs=["bold"]),
                     cli.col("({} instance{})".format(len(instances), "" if len(instances) == 1 else "s"), attrs=["dark"])))
        default = None
        for instance in instances:
            if not instance[0]:
                default = instance[1]
        for instance in instances:
            if instance[0] and instance[1] is default:
                state = cli.col("-> "+instance[1].name, "magenta")
            else:
                state = cli.col(_sname(instance[1]), **_sstyle(instance[1]))+_counts(cli, instance[1])
            data.append(("  "+ServiceHub.servicename(service, *instance[0]), state, _excinfo(instance[1], args.verbose)))
    Table(data).print(cli)


def _serviceCompleter(prefix, parsed_args, **kwargs):
    from services import ServiceHub
    available = set(ServiceHub.services())
    for service in parsed_args.services:
        available.discard(service)
    return available


def _cliSetupServiceParser(subp: ArgumentParser):
    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    load = sub.add_parser("load", help="(Re)load service(s)")
    load.set_defaults(_handle=cliServiceLoad)
    load.add_argument("-r", "--reload", action="store_true", help="Force reloading")
    load.add_argument("service", help="Service to (re)load").completer = _serviceCompleter
    load.add_argument("args", nargs="*", help="Service parameters")
    status = sub.add_parser("status", help="List services")
    status.set_defaults(_handle=cliServiceStatus)
    status.add_argument("-v", "--verbose", action="count", help="Show more information")
    status.add_argument("services", nargs="*", help="Only show status of specified service(s)").completer = _serviceCompleter


@Cli.command("service", _cliSetupServiceParser)
def cliServices(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from collections import defaultdict
from io import BytesIO


class AutoClean:
    """Simple context manager calling a function on exit."""

    def __init__(self, func, *args, **kwargs):
        """Initialize context manager

        Parameters
        ----------
        func : function
            Function to call on exit
        *args : tuple
            Arguments for func
        **kwargs : dict
            Keyword arguments for func
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        """Dummy method to allow use of `with` statement."""
        return self

    def __exit__(self, type, value, traceback):
        """Call function."""
        if self.func is not None:
            self.func(*self.args, **self.kwargs)

    def release(self):
        self.func = None


def createMapping(iterable, key, value=lambda x: x):
    """Convert list of elements to dictionary.

    Parameters
    ----------
    iterable : Iterable
        List of elements to map
    key : function
        Function returning the key given an element of `iterable`
    value : function
        Function returning the value given an element of `iterable`

    Returns
    -------
    mapping : dict
        A dictionary mapping each key to a list of values
    """
    mapping = dict()
    for item in iterable:
        k = key(item)
        if k in mapping:
            mapping[k].append(value(item))
        else:
            mapping[k] = value(item)
    return mapping


class GenericObject:
    def __init__(self, **attrs):
        for key, value in attrs.items():
            setattr(self, key, value)

    def __repr__(self):
        return "GenericObject({})".format(", ".join((key+"="+repr(getattr(self, key))
                                                     for key in dir(self) if not key.startswith("_"))))

    def __contains__(self, key):
        return key in dir(self)

    def __getitem__(self, item):
        return getattr(self, item)


def setDirectoryOwner(path, uid=None, gid=None):
    """Recursively set directory ownership of path.

    If neither uid nor gid is set, the function returns immediately without touching any files.

    Parameters
    ----------
    path : str
        Name of the target directory or file
    uid : str or in, optional
        uid of the new owner
    gid : str or int, optional
    """
    import os
    import shutil
    uid = None if uid == "" else uid
    gid = None if gid == "" else gid
    if uid is None and gid is None:
        return
    if os.path.isfile(path):
        shutil.chown(path, uid, gid)
    for path, subdirs, files in os.walk(path):
        shutil.chown(path, uid, gid)
        for entry in subdirs+files:
            shutil.chown(os.path.join(path, entry), uid, gid)


def setDirectoryPermission(path, mode):
    """Recursively set directory permissions of path.

    If mode is not set, the function returns immediately without touching any files.

    Parameters
    ----------
    path : str
        Name of the target directory or file
    uid : str or in, optional
        uid of the new owner
    gid : str or int, optional
    """
    import os
    if not mode:
        return
    if isinstance(mode, str):
        mode = int(mode, 0)
    if os.path.isfile(path):
        os.chmod(path, mode)
    dirmode = mode
    for field in range(0, 9, 3):
        dirmode |= 1<<field if mode & 7<<field else 0
    for path, subdirs, files in os.walk(path):
        os.chmod(path, dirmode)
        for entry in subdirs:
            os.chmod(os.path.join(path, entry), dirmode)
        for entry in files:
            os.chmod(os.path.join(path, entry), mode)

#######################################################
#
# Shamelessly stolen from `phpserialize` project and
# enhanced to support objects and classes out of the
# box.
#
#######################################################
def loadPSO(data, charset='utf-8', decode_strings=False,
         object_hook=None, array_hook=None):
    """Read a string from the open file object `fp` and interpret it as a
    data stream of PHP-serialized objects, reconstructing and returning
    the original object hierarchy.

    `fp` must provide a `read()` method that takes an integer argument.  Both
    method should return strings.  Thus `fp` can be a file object opened for
    reading, a `StringIO` object (`BytesIO` on Python 3), or any other custom
    object that meets this interface.

    `load` will read exactly one object from the stream.  See the docstring of
    the module for this chained behavior.

    If an object hook is given object-opcodes are supported in the serilization
    format.  The function is called with the class name and a dict of the
    class data members.  The data member names are in PHP format which is
    usually not what you want.  The `simple_object_hook` function can convert
    them to Python identifier names.

    If an `array_hook` is given that function is called with a list of pairs
    for all array items.  This can for example be set to
    `collections.OrderedDict` for an ordered, hashed dictionary.
    """
    fp = BytesIO(data)
    if array_hook is None:
        array_hook = dict

    def _expect(e):
        v = fp.read(len(e))
        if v != e:
            raise ValueError('failed expectation, expected %r got %r' % (e, v))

    def _read_until(delim):
        buf = []
        while 1:
            char = fp.read(1)
            if char == delim:
                break
            elif not char:
                raise ValueError('unexpected end of stream')
            buf.append(char)
        return b''.join(buf)

    def _load_array():
        items = int(_read_until(b':')) * 2
        _expect(b'{')
        result = []
        last_item = Ellipsis
        for idx in range(items):
            item = _unserialize()
            if last_item is Ellipsis:
                last_item = item
            else:
                result.append((last_item, item))
                last_item = Ellipsis
        _expect(b'}')
        return result

    def _load_class():
        unused = int(_read_until(b':'))
        _expect(b'{')
        data = _unserialize()
        _expect(b'}')
        return data

    def _unserialize():
        type_ = fp.read(1).lower()
        if type_ == b'n':
            _expect(b';')
            return None
        if type_ in b'idb':
            _expect(b':')
            data = _read_until(b';')
            if type_ == b'i':
                return int(data)
            if type_ == b'd':
                return float(data)
            return int(data) != 0
        if type_ == b's':
            _expect(b':')
            length = int(_read_until(b':'))
            _expect(b'"')
            data = fp.read(length)
            _expect(b'"')
            if decode_strings:
                data = data.decode(charset)
            _expect(b';')
            return data
        if type_ == b'a':
            _expect(b':')
            return array_hook(_load_array())
        if type_ in b'oc':
            _expect(b':')
            name_length = int(_read_until(b':'))
            _expect(b'"')
            name = fp.read(name_length)
            _expect(b'":')
            if decode_strings:
                name = name.decode(charset)
            return {name: dict(_load_array()) if type_ == b'o' else _load_class()}
        raise ValueError('unexpected opcode')

    return _unserialize()


class RecursiveDict(dict):
    """dict extension to handle keys in dottet notation."""

    def __init__(self, data=None, reconstruct=True):
        """Construct from data

        Parameters
        ----------
        data : dict, optional
            Dict to create from. The default is None.
        reconstruct : bool, optional
            Whether to cast contained dicts to RecursiveDict
        """
        if data:
            for key, value in data.items():
                self.insert(key, RecursiveDict(value) if reconstruct and isinstance(value, dict) else value)

    def flat(self):
        """Flatten RecursiveDicts to normal dict with dot notation.

        Returns
        -------
        data : dict
            Converted dict.
        """
        data = {}
        for key, value in self.items():
            if not isinstance(value, RecursiveDict):
                data[key] = value
                continue
            for k, v in value.flat().items():
                data[key if k is None else key+"."+k] = v
        return data

    def insert(self, key, value):
        """Insert key, creating sub-levels as necessary.

        Parameters
        ----------
        key : any
            Key of the value to insert
        value : any
            Value to insert
        """
        if not isinstance(key, str) or "." not in key:
            if isinstance(self.get(key), RecursiveDict):
                self[key][None] = value
            else:
                self[key] = value
            return
        pre, post = key.split(".", 1)
        if pre not in self:
            self[pre] = RecursiveDict()
        elif not isinstance(self[pre], RecursiveDict):
            self[pre] = RecursiveDict({None: self[pre]})
        self[pre].insert(post, value)

    def update(self, E=None, **F):
        """Override for dict.update, using insert for each key.

        Parameters
        ----------
        E : dict, optional
            DESCRIPTION. The default is None.
        **F : TYPE
            DESCRIPTION.
        """
        if isinstance(E, RecursiveDict):
            for key, value in E.items():
                if isinstance(self.get(key), dict) and isinstance(value, dict):
                    self[key].update(value)
                else:
                    self[key] = value
        elif isinstance(E, dict):
            for key, value in E.items():
                self.insert(key, value)
        for key, value in F.items():
            self.insert(key, value)
        return self


# Copied from the jellyfish package and slightly modified
def damerau_levenshtein_distance(s1, s2):
    """Compute Damerau-Levenshtein string distance

    Parameters
    ----------
    s1 : str
        First string
    s2 : str
        Second string

    Returns
    -------
    int
        Distance between the two strings
    """
    if not isinstance(s1, str) or not isinstance(s2, str):
        raise TypeError("Arguments must be string")

    len1 = len(s1)
    len2 = len(s2)
    infinite = len1 + len2

    # character array
    da = defaultdict(int)

    # distance matrix
    score = [[0] * (len2 + 2) for x in range(len1 + 2)]

    score[0][0] = infinite
    for i in range(0, len1 + 1):
        score[i + 1][0] = infinite
        score[i + 1][1] = i
    for i in range(0, len2 + 1):
        score[0][i + 1] = infinite
        score[1][i + 1] = i

    for i in range(1, len1 + 1):
        db = 0
        for j in range(1, len2 + 1):
            i1 = da[s2[j - 1]]
            j1 = db
            cost = 1
            if s1[i - 1] == s2[j - 1]:
                cost = 0
                db = j

            score[i + 1][j + 1] = min(
                score[i][j] + cost,
                score[i + 1][j] + 1,
                score[i][j + 1] + 1,
                score[i1][j1] + (i - i1 - 1) + 1 + (j - j1 - 1),
            )
        da[s1[i - 1]] = i

    return score[len1 + 1][len2 + 1]



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

import time


def gcToValue(value: int) -> int:
    return int.from_bytes(value.to_bytes(8, "big")[:6], "little")


def valueToGc(value: int) -> bytes:
    """-> rop_util.c:57."""
    return value.to_bytes(6, "little")


def makeEid(replid: int, gc: bytes) -> int:
    """-> rop_util.c:93."""
    return replid | (int.from_bytes(gc, "big") << 16)


def makeEidEx(replid: int, value: int) -> int:
    """-> rop_util.c:109."""
    gc = valueToGc(value)
    return makeEid(replid, gc)


def ntTime(timestamp=None) -> int:
    """Convert UNIX timestamp to Windows timestamp.

    Parameters
    ----------
    timestamp : float, optional
        Timestamp to convert. If not specified, use current time. The default is None.

    Returns
    -------
    int
        Windows NT timestamp
    """
    timestamp = timestamp or time.time()
    timestamp += 11644473600
    timestamp *= 10000000
    return int(timestamp)


def nxTime(timestamp) -> float:
    timestamp /= 10000000
    timestamp -= 11644473600
    return timestamp



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, CliError, ArgumentParser
from .common import proptagCompleter


def _runParserSetup(subp: ArgumentParser):
    subp.add_argument("--debug", "-d", action="store_true", help="Run in debug mode")
    subp.add_argument("--ip", "-i", default="::", type=str, help="Host address to bind to")
    subp.add_argument("--no-config-check", action="store_true", help="Skip configuration check")
    subp.add_argument("--port", "-p", default=5001, type=int, help="Host port to bind to")
    subp.add_argument("--tasq", "-t", action="store_true", help="Start TasQ server")


@Cli.command("run", _runParserSetup, help="Run the REST API")
def cliRun(args):
    cli = args._cli
    if cli.mode == "adhoc":
        raise CliError("Cannot run in adhoc mode")
    from tools import config
    error = config.validate()
    if error:
        cli.print(cli.col("Invalid configuration found: "+error, "yellow" if args.no_config_check else "red"))
        if not args.no_config_check:
            return 1
    from tools.tasq import TasQServer
    if args.tasq:
        TasQServer.start()
    from api.core import API
    import endpoints
    import importlib
    for group in endpoints.__all__:
        importlib.import_module("endpoints."+group)
    API.run(host=args.ip, port=args.port, debug=args.debug)
    TasQServer.stop()


def _versionParserSetup(subp: ArgumentParser):
    subp.add_argument("--api", "-a", action="store_true", help="Print API version")
    subp.add_argument("--backend", "-b", action="store_true", help="Print Backend version")
    subp.add_argument("--combined", "-c", action="store_true", help="Print combined version")


@Cli.command("version", _versionParserSetup, help="Show version information")
def cliVersion(args):
    from api import backendVersion, apiVersion
    cli = args._cli
    if args.api:
        cli.print(apiVersion)
    if args.backend:
        cli.print(backendVersion)
    if args.combined or not any((args.api, args.backend, args.combined)):
        vdiff = int(backendVersion.rsplit(".", 1)[1])-int(apiVersion.rsplit(".", 1)[1])
        if vdiff == 0:
            cli.print(apiVersion)
        else:
            cli.print("{}{:+}".format(apiVersion, vdiff))


def _setupTaginfo(subp: ArgumentParser):
    tagID = subp.add_argument("tagID", nargs="+", help="Numeric tag ID in decimal or hexadecimal or tag name glob")
    tagID.completer = proptagCompleter


@Cli.command("taginfo", _setupTaginfo, help="Print information about proptags")
def cliTaginfo(args):
    def printTag(ID):
        propname = PropTags.lookup(ID, "unknown")
        typename = PropTypes.lookup(ID, "unknown")
        proptype = cli.col("{:04x}".format(ID % (1 << 16)), attrs=["dark"])
        cli.print("0x{:04x}{} ({}): {}, type {}".format(ID >> 16, proptype, ID, propname, typename))

    cli = args._cli
    from tools.constants import PropTags, PropTypes
    PropTags.lookup(0)
    for tagid in args.tagID:
        if "*" in tagid or "?" in tagid:
            import fnmatch
            import re
            pattern = re.compile(fnmatch.translate(tagid.upper()))
            for ID, name in PropTags._lookup.items():
                if isinstance(ID, int) and pattern.match(name):
                    printTag(ID)
            continue
        try:
            ID = int(tagid, 0)
        except Exception:
            ID = getattr(PropTags, tagid.upper(), None)
            if ID is None or type(ID) != int:
                cli.print("Unknown tag '{}'".format(tagid))
                continue
        printTag(ID)


def _setupCliShell(subp: ArgumentParser):
    subp.description = "Start shell to process multiple CLI calls in a single session"
    subp.add_argument("-d", "--debug", action="store_true", help="Print more information")
    subp.add_argument("-x", "--exit", action="store_true", help="Exit on error")
    subp.add_argument("-n", "--no-history", action="store_true", help="Disable typed history")


def _historyDir():
    import os
    filePath = os.environ.get("XDG_CONFIG_HOME") or os.environ.get("HOME")+"/.config"
    return filePath+"/grommunio-admin"


def _loadHistory(args):
    cli = args._cli
    import os
    import readline
    if not args.no_history:
        filePath = os.path.expanduser("~/.grommunio-admin.history")
        try:
            readline.read_history_file(filePath)
            if args.debug:
                cli.print(cli.col("Loaded {} history entries".format(readline.get_current_history_length()), attrs=["dark"]))
            os.unlink(filePath)
            return
        except Exception as err:
            if args.debug:
                cli.print(cli.col("Failed to read legacy history file '{}': {}"
                                  .format(filePath, type(err).__name__ + " - ".join(str(arg) for arg in err.args)),
                                  attrs=["dark"]))
        filePath = _historyDir()+"/history"
        try:
            readline.read_history_file(filePath)
            if args.debug:
                cli.print(cli.col("Loaded {} history entries".format(readline.get_current_history_length()), attrs=["dark"]))
        except Exception as err:
            if args.debug:
                cli.print(cli.col("Failed to read history file '{}': {}"
                                  .format(filePath, type(err).__name__ + " - ".join(str(arg) for arg in err.args)),
                                  attrs=["dark"]))


@Cli.command("shell", _setupCliShell, help="Start interactive shell")
def cliShell(args):
    cli = args._cli

    def rlEnable(state):
        if rlAvail:
            readline.set_completer(cli.complete if state else lambda *args: None)
            readline.set_auto_history(state)

    if cli.stdin is None:
        cli.print(cli.col("No input available - aborting", "red"))
        return -1

    import shlex
    import sys
    from services import ServiceUnavailableError
    rlAvail = False
    interactive = cli.stdin.isatty()
    if interactive:
        cli.print("\x1b]2;grommunio-admin\x07", end="")
        if cli.host is None:
            cli.print("grommunio-admin shell. Type exit or press CTRL+D to exit.")
        else:
            cli.print("Starting remote admin shell. Type exit or press CTRL+D to exit.")
        try:
            import readline
            readline.set_completer_delims("")
            readline.parse_and_bind("tab: complete")
            readline.set_history_length(1000)
            _loadHistory(args)
            rlAvail = True
        except Exception as err:
            if args.debug:
                cli.print(cli.col("Failed to initialize readline: "+type(err).__name__ +
                                  " - ".join(str(arg) for arg in err.args), attrs=["dark"]))
            cli.print("Install readline module to enable autocompletion")
    elif cli.mode == "standalone":
        cli.print(cli.col("WARNING: The CLI is still under development and subject to changes."
                          " Be careful when using scripts.\n", "yellow"), file=sys.stderr)
    try:
        while True:
            rlEnable(True)
            try:
                prompt = ""
                if interactive:
                    prompt += cli.col("grommunio-admin", "cyan", attrs=["bold", "dark"])
                    prompt += "@"+cli.col(cli.host, "red", attrs=["bold", "dark"]) if cli.host is not None else ""
                    prompt += "> "
                command = cli.input(prompt).strip()
            except KeyboardInterrupt:
                cli.print("^C")
                continue
            if command == "":
                continue
            elif command == "exit":
                break
            try:
                rlEnable(False)
                result = cli.execute(shlex.split(command), secure=False) or 0
                if result != 0 and args.exit:
                    return result
            except SystemExit:
                pass
            except KeyboardInterrupt:
                cli.print("^C")
                continue
            except EOFError:
                raise
            except ServiceUnavailableError as err:
                cli.print(cli.col(err.args[0], "red"))
            except BaseException as err:
                if isinstance(err, AttributeError) and "_argcomplete_namespace" in err.args[0]:
                    continue
                if args.debug:
                    import traceback
                    cli.print(cli.col(traceback.format_exc(), attrs=["dark"]))
                cli.print(cli.col("An exception occurred ({}): {}".format(type(err).__name__,
                                                                         " - ".join(str(arg) for arg in err.args)),
                                  "red"))
    except EOFError:
        cli.print()
    if rlAvail and not args.no_history:
        try:
            import os
            filePath = _historyDir()
            os.makedirs(filePath, exist_ok=True)
            readline.write_history_file(filePath+"/history")
        except Exception as err:
            if args.debug:
                cli.print(cli.col("Failed to write history file: "+type(err).__name__ +
                                  " - ".join(str(arg) for arg in err.args), attrs=["dark"]))


@Cli.command("shrek")
def cliShrek(args):
    cli = args._cli
    cli.print(cli.col("⢀⡴⠑⡄⠀⠀⠀⠀⠀⠀⠀⣀⣀⣤⣤⣤⣀⡀\n"
                      "⠸⡇⠀⠿⡀⠀⠀⠀⣀⡴⢿⣿⣿⣿⣿⣿⣿⣿⣷⣦⡀\n"
                      "⠀⠀⠀⠀⠑⢄⣠⠾⠁⣀⣄⡈⠙⣿⣿⣿⣿⣿⣿⣿⣿⣆\n"
                      "⠀⠀⠀⠀⢀⡀⠁⠀⠀⠈⠙⠛⠂⠈⣿⣿⣿⣿⣿⠿⡿⢿⣆\n"
                      "⠀⠀⠀⢀⡾⣁⣀⠀⠴⠂⠙⣗⡀⠀⢻⣿⣿⠭⢤⣴⣦⣤⣹⠀⠀⠀⢀⢴⣶⣆\n"
                      "⠀⠀⢀⣾⣿⣿⣿⣷⣮⣽⣾⣿⣥⣴⣿⣿⡿⢂⠔⢚⡿⢿⣿⣦⣴⣾⠁⠸⣼⡿\n"
                      "⠀⢀⡞⠁⠙⠻⠿⠟⠉⠀⠛⢹⣿⣿⣿⣿⣿⣌⢤⣼⣿⣾⣿⡟⠉\n"
                      "⠀⣾⣷⣶⠇⠀⠀⣤⣄⣀⡀⠈⠻⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡇\n"
                      "⠀⠉⠈⠉⠀⠀⢦⡈⢻⣿⣿⣿⣶⣶⣶⣶⣤⣽⡹⣿⣿⣿⣿⡇\n"
                      "⠀⠀⠀⠀⠀⠀⠀⠉⠲⣽⡻⢿⣿⣿⣿⣿⣿⣿⣷⣜⣿⣿⣿⡇\n"
                      "⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⣷⣶⣮⣭⣽⣿⣿⣿⣿⣿⣿⣿\n"
                      "⠀⠀⠀⠀⠀⠀⣀⣀⣈⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠇\n"
                      "⠀⠀⠀⠀⠀⠀⢿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠃\n"
                      "⠀⠀⠀⠀⠀⠀⠀⠹⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡿⠟⠁\n"
                      "⠀⠀⠀⠀⠀⠀⠀⠀⠀⠉⠛⠻⠿⠿⠿⠿⠛⠉", "green"))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 grommunio GmbH

from dns import resolver, reversename
import socket

from .config import Config


def getHostByName(domain):
    try:
        # If host can be resolved
        socket.getaddrinfo(domain, None)
        return True
    except Exception:
        pass
    # Host could not be resolved
    return False


def getLocalIp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        s.connect((Config["dns"]["dudIP"], 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    s.close()
    return IP


externalResolver = resolver.Resolver()
externalResolver.nameservers = Config["dns"]["externalResolvers"]


def fullDNSCheck(domain: str):
    localIp = getLocalIp()
    externalIp = checkMyIP()
    mxRecords = checkMX(domain)
    autodiscover = checkAutodiscover(domain)
    autodiscoverSRV = checkAutodiscoverSRV(domain)
    autoconfig = checkAutoconfig(domain)
    txt = checkTXT(domain)
    dkim = checkDKIM(domain)
    dmarc = checkDMARC(domain)
    srv = checkAllSRV(domain)
    caldavTXT = checkCaldavTxt(domain)
    carddavTXT = checkCarddavTxt(domain)
    return {
        "localIp": localIp,
        "externalIp": externalIp,
        "mxRecords": mxRecords,
        "autodiscover": autodiscover,
        "autodiscoverSRV": autodiscoverSRV,
        "autoconfig": autoconfig,
        "txt": txt,
        "dkim": dkim,
        "dmarc": dmarc,
        "caldavTXT": caldavTXT,
        "carddavTXT": carddavTXT,
        **srv
    }


def checkMyIP():
    customResolver = resolver.Resolver()
    customResolver.nameservers = ["208.67.222.222", "208.67.220.220", "208.67.222.220"]
    res = None
    try:
        dnsAnswer = customResolver.query("myip.opendns.com")
        res = ", ".join([str(a) for a in dnsAnswer])
    except Exception:
        pass
    return res


def ip(domain: str):
    res = None
    try:
        dnsAnswer = resolver.query(domain)
        res = ", ".join([str(a) for a in dnsAnswer])
    except Exception:
        pass
    return res


def checkMX(domain: str):
    res = {
        "internalDNS": None,
        "externalDNS": None,
        "reverseLookup": None,
        "mxDomain": None,
    }
    try:
        mxRecords = resolver.query(domain, "MX")
        mxDomain = mxRecords[0].exchange # Mail-domain of domain
        res["mxDomain"] = str(mxDomain)
        try:
            mxResolved = externalResolver.query(mxDomain, "A") # IP of mail-domain
            res["externalDNS"] = ", ".join([str(r) for r in mxResolved])

            # Reverse lookup
            addresses = [reversename.from_address(str(r)) for r in mxResolved]
            res["reverseLookup"] = str(resolver.query(addresses[0], "PTR")[0])
        except Exception:
            pass
        try:
            mxResolved = resolver.query(mxDomain, "A")
            res["internalDNS"] = ", ".join([str(r) for r in mxResolved])
        except Exception:
            pass
    except Exception:
        pass
    return res


def checkAutodiscover(domain: str):
    return defaultDNSQuery("autodiscover.", domain)


def checkAutoconfig(domain: str):
    return defaultDNSQuery("autoconfig.", domain)


def checkAllSRV(domain: str):
    res = {f"{subdomain}SRV": defaultDNSQuery(f"_{subdomain}._tcp.", domain, recordType="SRV")
           for subdomain in ["submission", "imap", "imaps", "pop3", "pop3s", "caldav", "caldavs", "carddav", "carddavs"]}
    return res


def checkAutodiscoverSRV(domain: str):
    res = None
    resExternal = None
    adIp = None
    try:
        records = resolver.query("_autodiscover._tcp." + domain, "SRV")
        res = ", ".join([str(r) for r in records])
    except Exception:
        pass
    try:
        records = externalResolver.query("_autodiscover._tcp." + domain, "SRV")
        resExternal = ", ".join([str(r) for r in records])
        adIp = ip(str(records[0]).split(" ")[3])
    except Exception:
        pass
    return {"internalDNS": res, "externalDNS": resExternal, "ip": adIp }


def checkTXT(domain: str):
    res = None
    resExternal = None
    try:
        txtRecords = resolver.query(domain, "TXT")
        res = ", ".join([str(r) for r in txtRecords if str(r).startswith('"v=spf1')])
    except Exception:
        pass

    try:
        txtRecordsExternal = externalResolver.query(domain, "TXT")
        resExternal = ", ".join([str(r) for r in txtRecordsExternal if str(r).startswith('"v=spf1')])
    except Exception:
        pass
    return {"internalDNS": res, "externalDNS": resExternal}


def checkDKIM(domain: str):
    return defaultDNSQuery("dkim._domainkey.", domain, recordType="TXT")


def checkDMARC(domain: str):
    return defaultDNSQuery("_dmarc.", domain, recordType="TXT")


def checkCaldavTxt(domain: str):
    return defaultDNSQuery("_caldavs._tcp.", domain, recordType="TXT")


def checkCarddavTxt(domain: str):
    return defaultDNSQuery("_carddavs._tcp.", domain, recordType="TXT")


def defaultDNSQuery(subdomain: str, domain: str, recordType="A", path=""):
    res = None
    resExternal = None
    try:
        records = resolver.query(subdomain + domain + path, recordType)
        res = ", ".join([str(r) for r in records])
    except Exception:
        pass

    try:
        records = externalResolver.query(subdomain + domain + path, recordType)
        resExternal = ", ".join([str(r) for r in records])
    except Exception:
        pass
    return {"internalDNS": res, "externalDNS": resExternal}



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from . import defaultListQuery

from flask import jsonify, request

from tools.permissions import DomainAdminROPermission, SystemAdminPermission, SystemAdminROPermission
from tools.tasq import TasQServer


@API.route(api.BaseRoute+"/tasq/status", methods=["GET"])
@secure(requireAuth=False)
def getTasQStatus():
    return jsonify(running=TasQServer.running(),
                   queued=TasQServer.queued(),
                   workers=TasQServer.workers())


@API.route(api.BaseRoute+"/tasq/start", methods=["POST"])
@secure()
def startTasQ():
    checkPermissions(SystemAdminPermission())
    TasQServer.start(int(request.args["procs"]) if "procs" in request.args else None)
    return jsonify(message="TasQ server started")


@API.route(api.BaseRoute+"/tasq/stop", methods=["POST"])
@secure()
def stopTasQ():
    checkPermissions(SystemAdminPermission())
    TasQServer.stop(float(request.args["timeout"]) if "timeout" in request.args else None)
    return jsonify(message="TasQ server stopped")


@API.route(api.BaseRoute+"/tasq/tasks", methods=["GET"])
@secure(requireDB=102, authLevel="user")
def getTasQTasks():
    from orm.misc import TasQ
    tasks = defaultListQuery(TasQ, result="list")
    userPerms = request.auth["user"].permissions()
    nofilter = SystemAdminROPermission() in userPerms
    verbosity = int(request.args.get("level", 1))
    data = [task.todict(verbosity) for task in tasks if nofilter or task.permission in userPerms]
    return jsonify(data=data)


@API.route(api.BaseRoute+"/tasq/tasks/<int:ID>", methods=["GET"])
@secure(requireDB=102, authLevel="user")
def getTasQTask(ID):
    from orm.misc import TasQ
    task = TasQ.query.filter(TasQ.ID == ID).first()
    if task is None:
        return jsonify(message="Task not found"), 404
    checkPermissions(task.permission)
    return jsonify(task.todict(int(request.args.get("level", 2))))


@API.route(api.BaseRoute+"/tasq/notify", methods=["POST"])
@secure(requireAuth=False)
def notifyTasQ():
    pulled = TasQServer.pull()
    return jsonify(message="Pulled {} task{} from the database".format(pulled, "" if pulled == 1 else "s"))
